import uuid
from typing import Any, Dict, Iterable, Optional, Union

import click
from typing_extensions import Literal

from gcs_cli.helpers import requires_login
from gcs_cli.parsing import (
    EXPLICIT_NULL,
    collection_create_and_update_params,
    globus_command,
)
from gcs_cli.prettyprint import FORMAT_TEXT_RECORD, formatted_print
from gcs_cli.sdk_client.data import (
    CollectionDocument,
    GoogleCloudStorageCollectionPolicies,
    POSIXCollectionPolicies,
    POSIXStagingCollectionPolicies,
)
from gcs_cli.services.gcs import get_client


@globus_command("create", short_help="Create a new Mapped Collection")
@click.argument("STORAGE_GATEWAY_ID")
@click.argument("BASE_PATH")
@collection_create_and_update_params(create=True)
@click.option(
    "--identity-id",
    help=(
        "Globus Auth identity to who acts as the owner of this Collection. This "
        "identity is an administrator on the Endpoint"
    ),
)
@requires_login
def collection_create_mapped(
    storage_gateway_id: uuid.UUID,
    base_path: str,
    identity_id: Optional[str],
    google_project_id: Optional[str],
    posix_sharing_group_allow: Optional[Iterable[str]] = None,
    posix_sharing_group_deny: Optional[Iterable[str]] = None,
    posix_staging_sharing_group_allow: Optional[Iterable[str]] = None,
    posix_staging_sharing_group_deny: Optional[Iterable[str]] = None,
    sharing_restrict_paths: Optional[Dict[str, Iterable[str]]] = None,
    sharing_users_allow: Optional[Iterable[str]] = None,
    sharing_users_deny: Optional[Iterable[str]] = None,
    user_message: Optional[str] = None,
    user_message_link: Optional[str] = None,
    verify: Union[Literal["force"], Literal["disable"], Literal["default"]] = None,
    **kwargs: Any
) -> None:
    """
    Create a new Mapped Collection, rooted on some given path within an
    existing Storage Gateway.

    The STORAGE_GATEWAY_ID is the Storage Gateway which will host the Collection.
    """
    if identity_id is not None:
        kwargs["identity_id"] = identity_id
    if google_project_id is not None:
        kwargs["policies"] = GoogleCloudStorageCollectionPolicies(
            project=google_project_id
        )

    if sharing_users_allow is not None:
        kwargs["sharing_users_allow"] = sharing_users_allow

    if sharing_users_deny is not None:
        kwargs["sharing_users_deny"] = sharing_users_deny

    if posix_sharing_group_allow is not None or posix_sharing_group_deny is not None:
        # Added in 5.4.8 for POSIX connector
        policies = {}
        if posix_sharing_group_allow is not None:
            policies["sharing_groups_allow"] = posix_sharing_group_allow

        if posix_sharing_group_deny is not None:
            policies["sharing_groups_deny"] = posix_sharing_group_deny
        kwargs["policies"] = POSIXCollectionPolicies(**policies)

    if (
        posix_staging_sharing_group_allow is not None
        or posix_staging_sharing_group_deny is not None
    ):
        # Added in 5.4.10 for POSIX staging connector
        policies = {}
        if posix_staging_sharing_group_allow is not None:
            policies["sharing_groups_allow"] = posix_staging_sharing_group_allow

        if posix_staging_sharing_group_deny is not None:
            policies["sharing_groups_deny"] = posix_staging_sharing_group_deny
        kwargs["policies"] = POSIXStagingCollectionPolicies(**policies)

    if sharing_restrict_paths is not None:
        if sharing_restrict_paths == EXPLICIT_NULL:
            kwargs["sharing_restrict_paths"] = None
        else:
            kwargs["sharing_restrict_paths"] = sharing_restrict_paths

    if user_message is not None:
        if user_message == EXPLICIT_NULL:
            kwargs["user_message"] = None
        else:
            kwargs["user_message"] = user_message

    if user_message_link is not None:
        if user_message_link == EXPLICIT_NULL:
            kwargs["user_message_link"] = None
        else:
            kwargs["user_message_link"] = user_message_link

    # These are added in GCS 5.4.4, so if neither --enable-https or
    # --disable-https are set, then we'll not try to set the value for
    # backward compatibility
    if kwargs.get("enable_https") and kwargs.get("disable_https"):
        raise click.exceptions.UsageError(
            "Only one of --enable-https and --disable-https may be used"
        )
    if kwargs.get("enable_https") is False:
        kwargs.pop("enable_https")

    if kwargs.get("disable_https"):
        kwargs["enable_https"] = False

    kwargs.pop("disable_https", None)

    if verify is not None:
        if verify.lower() == "force":
            kwargs["force_verify"] = True
            kwargs["disable_verify"] = False
        elif verify.lower() == "disable":
            kwargs["force_verify"] = False
            kwargs["disable_verify"] = True
        else:
            kwargs["force_verify"] = False
            kwargs["disable_verify"] = False

    collection_doc = CollectionDocument(
        storage_gateway_id=storage_gateway_id,
        collection_type="mapped",
        collection_base_path=base_path,
        **kwargs
    )
    client = get_client()
    res = client.create_collection(collection_doc)

    # output
    formatted_print(
        res, fields=[("Collection ID", "id")], text_format=FORMAT_TEXT_RECORD
    )
