from typing import Any, Callable, Dict, Iterable, List, Optional, TypeVar, cast

import click

from gcs_cli.parsing.types import (
    JSONStringOrFile,
    StringOrNull,
    nullable_multi_callback,
)

_T = TypeVar("_T")


def posix_options(create: bool = False) -> Callable[[_T], _T]:
    # see shared_options.storage_gateway_create_and_update_params for an explanation of
    # the need for these two variants
    multi_use_option_str = "Give this option multiple times"
    if not create:
        multi_use_option_str = "Give this option multiple times in a single command"

    def inner_decorator(f: _T) -> _T:
        # POSIX policies
        f = click.option(
            "--posix-group-allow",
            multiple=True,
            callback=nullable_multi_callback(""),
            help=(
                "POSIX group allowed access to this Storage Gateway. "
                "{} to allow multiple groups. ".format(multi_use_option_str)
                + 'Set a value of "" to clear this'
            ),
        )(f)
        f = click.option(
            "--posix-group-deny",
            multiple=True,
            callback=nullable_multi_callback(""),
            help=(
                "POSIX group denied access to this Storage Gateway. "
                "{} to deny multiple groups. ".format(multi_use_option_str)
                + 'Set a value of "" to clear this'
            ),
        )(f)
        return f

    return inner_decorator


def nullable_environment_callback(
    null: str = "null",
) -> Callable[
    [click.Context, click.Parameter, str], Optional[Iterable[Dict[str, str]]]
]:
    """
    Validate that non-empty strings in the multi-value environment list contain
    "="
    """

    def callback(
        ctx: click.Context, param: click.Parameter, value: str
    ) -> Optional[Iterable[Dict[str, str]]]:
        multi_value = nullable_multi_callback(null=null)(ctx, param, value)

        if not multi_value:
            return None

        newval: List[Dict[str, str]] = []
        for v in multi_value:
            res = v.split("=", 1)
            if len(res) != 2:
                raise click.BadParameter(
                    "Environment values must be in the format VAR=VALUE"
                )
            newval.append({"name": res[0], "value": res[1]})
        return newval

    return callback


def posix_staging_options(create: bool = False) -> Callable:
    # see shared_options.storage_gateway_create_and_update_params for an explanation of
    # the need for these two variants
    multi_use_option_str = "Give this option multiple times"
    if not create:
        multi_use_option_str = "Give this option multiple times in a single command"

    def inner_decorator(f: Callable) -> Callable:
        f = posix_options(create=create)(f)
        # POSIX Staging policies
        f = click.option(
            "--stage-app",
            help=("Path to the file staging app to use for this storage gateway"),
        )(f)
        f = click.option(
            "--environment",
            multiple=True,
            callback=nullable_environment_callback(""),
            metavar="NAME=VALUE",
            help=(
                "Variables to set in the environment when executing the stage_app. "
                "{} to set multiple environment variables. ".format(
                    multi_use_option_str
                )
                + 'Set a value of "" to clear this'
            ),
        )(f)
        return f

    return inner_decorator


def _s3_endpoint_option(f: _T, name: str) -> _T:
    return cast(
        _T,
        click.option(
            "--s3-endpoint", help="Region-specific URI of the {} API".format(name)
        )(f),
    )


def _bucket_option(f: _T, name: str) -> _T:
    connector_text = ""
    if name == "S3":
        connector_text = (
            "This is the only way to make public S3 buckets "
            "visible to the root of an S3 Storage Gateway. "
        )
    return cast(
        _T,
        click.option(
            "--bucket",
            multiple=True,
            callback=nullable_multi_callback(),
            help=(
                "Bucket to include in the root of the Storage Gateway. "
                "Give this option multiple times to include multiple "
                "buckets. "
            )
            + connector_text
            + 'Set a value of "null" to clear this',
        )(f),
    )


def _google_options(f: _T, name: str) -> _T:
    f = click.option(
        "--google-client-id",
        help=("Application Client Id registered with Google to access " + name),
    )(f)
    f = click.option(
        "--google-client-secret",
        help=(
            "Secret created by Google to access Google as "
            "the given --google-client-id"
        ),
    )(f)
    return f


def onedrive_options(f: _T) -> _T:
    f = _ms_options(f, "OneDrive")
    f = click.option(
        "--ms-user-api-rate-limit",
        type=int,
        help="OneDrive API rate limit. Number of requests per second per user",
    )(f)
    return f


def _ms_options(f: _T, name: str) -> _T:
    f = click.option(
        "--ms-client-id",
        help=("Application Client Id registered with Microsoft to access " + name),
    )(f)
    f = click.option(
        "--ms-client-secret",
        help=(
            "Secret created by Microsoft to access " + name + " as "
            "the given --ms-client-id"
        ),
    )(f)
    f = click.option(
        "--ms-tenant",
        help=(
            "Microsoft Tenant Id.  Required when application is configured "
            "in single-tenant mode."
        ),
        type=StringOrNull(),
    )(f)
    return f


def azure_blob_options(f: _T) -> _T:
    f = _ms_options(f, "Azure Blob Storage")
    f = click.option(
        "--azure-credential-type",
        type=click.Choice(["user", "service_principal"]),
        help=(
            "Set the type of credential used for authentication to Azure. "
            "'user' requires the user to register a credential by logging in "
            "to Azure AD.  'service_principal' uses the configured client_id "
            "and client_secret values to authenticate as a service principal for "
            "all users."
        ),
    )(f)
    f = click.option(
        "--azure-storage-account",
        help="Azure storage account with blob storage containers",
        type=StringOrNull(),
    )(f)
    f = click.option(
        "--azure-adls/--azure-no-adls",
        default=None,
        help=(
            "Indicate whether Azure Data Lake Gen2 hierarchical namespace is "
            "enabled on the storage account."
        ),
    )(f)
    return f


def blackpearl_options(f: _T) -> _T:
    f = _s3_endpoint_option(f, "BlackPearl")
    f = click.option(
        "--blackpearl-access-id-file", help="Path to the BlackPearl user mapping file"
    )(f)
    return f


def box_options(f: _T) -> _T:
    def validate_box_settings(
        ctx: click.Context, param: click.Parameter, value: Optional[Dict[str, Any]]
    ) -> Optional[Dict[str, Any]]:
        if value is None:
            return None
        if "enterpriseID" not in value:
            raise click.UsageError("--box-settings must contain enterpriseID")
        if "boxAppSettings" not in value:
            raise click.UsageError("--box-settings must contain boxAppSettings")
        return value

    return cast(
        _T,
        click.option(
            "--box-settings",
            callback=validate_box_settings,
            type=JSONStringOrFile(),
            help="The Box App Settings JSON data, as a string or a file",
        )(f),
    )


def ceph_options(f: _T) -> _T:
    f = _s3_endpoint_option(f, "Ceph")
    f = _bucket_option(f, "Ceph")
    f = click.option(
        "--ceph-admin-key-id",
        help="Key ID of an admin key used to resolve Ceph usernames to credentials",
    )(f)
    f = click.option(
        "--ceph-admin-secret-key",
        help="Secret key corresponding to --ceph-admin-key-id",
    )(f)
    return f


def google_drive_options(f: _T) -> _T:
    f = _google_options(f, "Google Drive")
    f = click.option(
        "--google-drive-user-api-rate-quota",
        type=int,
        help="Negotiated rate quota for Queries per 100 seconds per use",
    )(f)
    return f


def google_cloud_storage_options(f: _T) -> _T:
    f = _bucket_option(f, "Google Cloud Storage")
    f = _google_options(f, "Google Cloud Storage")
    f = click.option(
        "--google-cloud-storage-project",
        multiple=True,
        callback=nullable_multi_callback(),
        help=(
            "Project this Storage Gateway is allowed to access. "
            "Give this option multiple times to allow multiple "
            "projects. If no projects are specified, project access is "
            "unrestricted. "
            'Set a value of "null" to clear this'
        ),
    )(f)

    def validate_key(
        ctx: click.Context, param: click.Parameter, value: Dict[str, Any]
    ) -> Optional[Dict[str, Any]]:
        if value is None:
            return None
        if "project_id" not in value:
            raise click.UsageError(
                "--google-service-account-key must contain project_id"
            )
        if "client_id" not in value:
            raise click.UsageError(
                "--google-service-account-key must contain client_id"
            )
        return value

    f = click.option(
        "--google-service-account-key",
        callback=validate_key,
        type=JSONStringOrFile(),
        help=(
            "The Google service account key file, which will configure "
            "service account authentication.  When this is provided, all "
            "users will access Google Cloud Storage using these service "
            "account credentials."
        ),
    )(f)
    return f


def activescale_options(f: _T) -> _T:
    f = _s3_endpoint_option(f, "ActiveScale S3")
    f = _bucket_option(f, "ActiveScale S3")
    f = click.option(
        "--s3-user-credential/--s3-unauthenticated",
        default=None,
        help=(
            "Either require an ActiveScale S3 user credential to access this "
            "Storage Gateway or not. If unauthenticated, the Storage "
            "Gateway will only allow access to public buckets"
        ),
    )(f)
    return f


def s3_options(f: _T) -> _T:
    f = _s3_endpoint_option(f, "S3")
    f = _bucket_option(f, "S3")
    f = click.option(
        "--s3-user-credential/--s3-unauthenticated",
        default=None,
        help=(
            "Either require an S3 user credential to access this "
            "Storage Gateway or not. If unauthenticated, the Storage "
            "Gateway will only allow access to public S3 buckets"
        ),
    )(f)
    return f


def irods_options(f: _T) -> _T:
    f = click.option(
        "--irods-environment-file",
        default=None,
        help=(
            "Path to iRODS environent file on the endpoint.  "
            "Must be readable by user gcsweb."
        ),
    )(f)
    f = click.option(
        "--irods-authentication-file",
        default=None,
        help=(
            "Path to iRODS authentication file on the endpoint, containing valid "
            "credentials for the irods admin user defined in the environment file.  "
            "Must be readable by user gcsweb."
        ),
    )(f)
    return f
