import logging
import uuid
from typing import Any, Dict, Iterable, Union

from globus_sdk import GCSAPIError, GlobusHTTPResponse, paging
from globus_sdk.authorizers import (
    AccessTokenAuthorizer,
    ClientCredentialsAuthorizer,
    GlobusAuthorizer,
    RefreshTokenAuthorizer,
)
from globus_sdk.client import BaseClient
from globus_sdk.services.gcs import IterableGCSResponse, UnpackingGCSResponse

from gcs_cli.sdk_client.host_header_ssl_adapter import HostHeaderSSLAdapter

log = logging.getLogger(__name__)


class GCSClient(BaseClient):
    """
    Client for the Globus Connect Server API that is run separately by
    each Globus Connect Server instance starting at GCSv5.

    This class provides helper methods for most common resources in the
    GCS API, and the common low-level interface from
    :class:`BaseClient <globus_sdk.base.BaseClient>` of ``get``, ``put``,
    ``post``, ``patch`` and ``delete`` methods,
    which can be used to access any API resource.

    **Parameters**

    ``gcs_fdqn`` (*str*)
    The fully qualified domain name of the GCS instance this client will
    make calls to

    ``authorizer`` (*:class:`GlobusAuthorizer \
                        <globus_sdk.authorizers.GlobusAuthorizer>`*)
    An authorizer instance used for calls to the GCS API

    **Paginated Calls**

    Methods which support pagination can be called as paginated or unpaginated methods.
    If the method name is ``foo``, the paginated version is
    ``paginated.foo``. e.g. ``GCSClient.get_storage_gateway_list`` and
    ``GCSClient.paginated.get_storage_gateway_list`` are both valid.

     **Methods**

     *  :py:meth:`.set_address`
     *  :py:meth:`.get_storage_gateway_list`
     *  :py:meth:`.create_storage_gateway`
     *  :py:meth:`.get_storage_gateway`
     *  :py:meth:`.update_storage_gateway`
     *  :py:meth:`.delete_storage_gateway`
     *  :py:meth:`.get_endpoint`
     *  :py:meth:`.update_endpoint`
     *  :py:meth:`.get_endpoint_domain`
     *  :py:meth:`.delete_endpoint_domain`
     *  :py:meth:`.put_endpoint_domain`
     *  :py:meth:`.get_collection_list`
     *  :py:meth:`.create_collection`
     *  :py:meth:`.get_collection`
     *  :py:meth:`.put_collection_domain`
     *  :py:meth:`.get_collection_domain`
     *  :py:meth:`.delete_collection_domain`
     *  :py:meth:`.update_collection`
     *  :py:meth:`.delete_collection`
     *  :py:meth:`.get_node_list`
     *  :py:meth:`.get_node`
     *  :py:meth:`.update_node`
     *  :py:meth:`.delete_node`
     *  :py:meth:`.create_sharing_policy`
     *  :py:meth:`.get_sharing_policy_list`
     *  :py:meth:`.get_sharing_policy`
     *  :py:meth:`.delete_sharing_policy`
    """

    service_name = "gcs"

    # disallow basic auth
    allowed_authorizer_types = [
        AccessTokenAuthorizer,
        RefreshTokenAuthorizer,
        ClientCredentialsAuthorizer,
    ]
    error_class = GCSAPIError

    def __init__(
        self, gcs_fqdn: str, authorizer: GlobusAuthorizer = None, **kwargs: Any
    ):
        BaseClient.__init__(
            self,
            base_url="https://{}/api/".format(gcs_fqdn),
            authorizer=authorizer,
            **kwargs,
        )
        self.gcs_fqdn = gcs_fqdn
        self.headers: Dict[str, str] = {}
        # setup the HostHeaderSSLAdapter in case explicit host is set
        self.transport.session.mount("https://", HostHeaderSSLAdapter())

    def set_address(self, address: str) -> None:
        """
        Set a host header. With use of HostHeaderSSLAdapter, this allows requests by IP
        to proceed with an explicit hostname.

        Can accept just the address (raw) or a full URL.
        """
        self.headers["Host"] = self.gcs_fqdn
        if address.startswith("https://"):
            self.base_url = address
        else:
            self.base_url = "https://{}/api/".format(address)

    #
    # Storage Gateways
    #

    @paging.has_paginator(
        paging.MarkerPaginator,
        items_key="data",
    )
    def get_storage_gateway_list(
        self, include: Iterable[str] = None, **params: Any
    ) -> IterableGCSResponse:
        """
        ``GET /storage_gateways``

        ** Parameters **

        ``include`` (*list of str*)
        List of document types to include in the response. If private_policies
        is in the list, then include private storage gateway policies in the
        attached storage_gateways document. This requires an administrator
        role on the Endpoint.
        """
        if include is not None:
            include = list(include)
            params = {**params, "include": include}

        path = "/storage_gateways"
        return IterableGCSResponse(
            self.get(path, query_params=params, headers=self.headers)
        )

    def create_storage_gateway(
        self, data: Dict[str, Any], **params: Any
    ) -> UnpackingGCSResponse:
        """
        ``POST /storage_gateways``

        ** Parameters **

        ``data`` (*dict*)
        Data in the format of a Storage Gateway document, it is recommended
        to use the `StorageGatewayDocumment` class to construct this data.
        """
        path = "/storage_gateways"
        return UnpackingGCSResponse(
            self.post(path, data=data, query_params=params, headers=self.headers),
            "storage_gateway",
        )

    def get_storage_gateway(
        self,
        storage_gateway_id: Union[uuid.UUID, str],
        include: Iterable[str] = None,
        **params: Any,
    ) -> UnpackingGCSResponse:
        """
        ``GET /storage_gateways/<storage_gateway_id>``

        ** Parameters **

        ``storage_gateway_id`` (*uuid or str*)
        UUID for the Storage Gateway to be gotten

        ``include`` (*list of str*)
        List of document types to include in the response. If private_policies
        is in the list, then include private storage gateway policies in the
        attached storage_gateways document. This requires an administrator
        role on the Endpoint.
        """
        storage_gateway_id = storage_gateway_id
        if include is not None:
            include = list(include)
            params = {**params, "include": include}

        path = f"/storage_gateways/{storage_gateway_id}"
        return UnpackingGCSResponse(
            self.get(path, query_params=params, headers=self.headers), "storage_gateway"
        )

    def update_storage_gateway(
        self, storage_gateway_id: Union[str, uuid.UUID], data: Dict, **params: Any
    ) -> GlobusHTTPResponse:
        """
        ``PATCH /storage_gateways/<storage_gateway_id>``

        ** Parameters **

        ``storage_gatweay_id`` (*str*)
        UUID for the Storage Gateway to be updated

        ``data`` (*dict*)
        Data in the format of a Storage Gateway document, it is recommended
        to use the `StorageGatewayDocumment` class to construct this data.
        """
        path = f"/storage_gateways/{storage_gateway_id}"
        return self.patch(path, data=data, query_params=params, headers=self.headers)

    def delete_storage_gateway(
        self, storage_gateway_id: Union[str, uuid.UUID], **params: Any
    ) -> GlobusHTTPResponse:
        """
        ``DELETE /storage_gateways/<storage_gateway_id>``

        ** Parameters **

        ``storage_gatweay_id`` (*uuid or str*)
        UUID for the Storage Gateway to be deleted
        """
        path = f"/storage_gateways/{storage_gateway_id}"
        return self.delete(path, query_params=params, headers=self.headers)

    #
    # Endpoint
    #

    def get_endpoint(self, **params: Any) -> UnpackingGCSResponse:
        """
        ``GET /endpoint``
        """
        return UnpackingGCSResponse(
            self.get("/endpoint", query_params=params, headers=self.headers), "endpoint"
        )

    def update_endpoint(self, data: Dict, **params: Any) -> GlobusHTTPResponse:
        """
        ``PATCH /endpoint``

        ** Parameters **

        ``data`` (*dict*)
          Data in the format of a partial Endpoint document.
        """
        return self.patch(
            "/endpoint", data=data, query_params=params, headers=self.headers
        )

    def get_endpoint_domain(self, **params: Any) -> UnpackingGCSResponse:
        """
        GET /endpoint/domain
        """
        log.info("GCSClient.get_endpoint_domain()")
        return UnpackingGCSResponse(
            self.get("/endpoint/domain", headers=self.headers), "domain"
        )

    def put_endpoint_domain(self, data: Dict, **params: Any) -> GlobusHTTPResponse:
        """
        PUT /endpoint/domain
        """
        log.info("GCSClient.put_endpoint_domain()")
        return self.put("/endpoint/domain", data=data, headers=self.headers)

    def delete_endpoint_domain(self, **params: Any) -> GlobusHTTPResponse:
        """
        DELETE /endpoint/domain
        """
        log.info("GCSClient.delete_endpoint_domain()")
        return self.delete("/endpoint/domain", headers=self.headers)

    def put_endpoint_owner(self, data: Dict, **params: Any) -> GlobusHTTPResponse:
        """
        PUT /endpoint/owner
        """
        log.info("GCSClient.put_endpoint_owner()")
        return self.put("/endpoint/owner", data=data, headers=self.headers)

    def put_endpoint_subscription_id(
        self, data: Dict, **params: Any
    ) -> GlobusHTTPResponse:
        """
        PUT /endpoint/subscription_id
        """
        log.info("GCSClient.put_endpoint_subscription_id()")
        return self.put("/endpoint/subscription_id", data=data, headers=self.headers)

    #
    # Collections
    #

    @paging.has_paginator(
        paging.MarkerPaginator,
        items_key="data",
    )
    def get_collection_list(
        self, include: str = None, **params: Any
    ) -> IterableGCSResponse:
        """
        GET /collections
        """
        path = "/collections"
        if include is not None:
            params = {**params, "include": include}
        return IterableGCSResponse(
            self.get(path, query_params=params, headers=self.headers)
        )

    def create_collection(self, data: Dict, **params: Any) -> UnpackingGCSResponse:
        """
        POST /collections

        ** Parameters **

        ``data`` (*dict*)
          Data in the format of a Collection document, it is recommended to use
          the `CollectionDocumment` class to construct this data.
        """
        path = "/collections"
        return UnpackingGCSResponse(
            self.post(path, data=data, query_params=params, headers=self.headers),
            "collection",
        )

    def get_collection(
        self, collection_id: Union[uuid.UUID, str], **params: Any
    ) -> UnpackingGCSResponse:
        """
        GET /collections/{collection_id}

        ** Parameters **

        ``collection_id`` (*str*)
        UUID for the Collection to be gotten
        """
        path = f"/collections/{collection_id}"
        return UnpackingGCSResponse(
            self.get(path, query_params=params, headers=self.headers), "collection"
        )

    def update_collection(
        self, collection_id: Union[uuid.UUID, str], data: Dict, **params: Any
    ) -> GlobusHTTPResponse:
        """
        PATCH /collections/{collection_id}

        ** Parameters **

        ``collection_id`` (*str*)
        UUID for the Collection to be updated

        ``data`` (*dict*)
        Data in the format of a Collection document, it is recommended to use
        the `CollectionDocumment` class to construct this data.
        """
        path = f"/collections/{collection_id}"
        return self.patch(path, data=data, query_params=params, headers=self.headers)

    def delete_collection(
        self, collection_id: Union[uuid.UUID, str], **params: Any
    ) -> GlobusHTTPResponse:
        """
        DELETE /collections/{collection_id}

        ** Parameters **

        ``collection_id`` (*str*)
        UUID for the Collection to be deleted
        """
        path = f"/collections/{collection_id}"
        return self.delete(path, query_params=params, headers=self.headers)

    def get_collection_domain(
        self, collection_id: Union[uuid.UUID, str], **params: Any
    ) -> UnpackingGCSResponse:
        """
        GET /collections/{collection_id}/domain
        """
        log.info("GCSClient.get_collection_domain()")
        path = f"/collections/{collection_id}/domain"
        return UnpackingGCSResponse(self.get(path, headers=self.headers), "domain")

    def put_collection_domain(
        self, collection_id: Union[uuid.UUID, str], data: Dict, **params: Any
    ) -> GlobusHTTPResponse:
        """
        PUT /collections/{collection_id}/domain
        """
        log.info("GCSClient.put_collection_domain()")
        path = f"/collections/{collection_id}/domain"
        return self.put(path, data=data, headers=self.headers)

    def delete_collection_domain(
        self, collection_id: Union[uuid.UUID, str], **params: Any
    ) -> GlobusHTTPResponse:
        """
        DELETE /collections/{collection_id}/domain
        """
        log.info("GCSClient.delete_collection_domain()")
        path = f"/collections/{collection_id}/domain"
        return self.delete(path, headers=self.headers)

    #
    # Nodes
    #

    @paging.has_paginator(
        paging.MarkerPaginator,
        items_key="data",
    )
    def get_node_list(self, **params: Any) -> IterableGCSResponse:
        """
        GET /nodes
        """
        path = "/nodes"
        return IterableGCSResponse(
            self.get(path, query_params=params, headers=self.headers)
        )

    def get_node(
        self, node_id: Union[uuid.UUID, str], **params: Any
    ) -> UnpackingGCSResponse:
        """
        GET /nodes/{node_id}

        ** Parameters **

        ``node_id`` (*str*)
          UUID for the Node to fetch
        """
        log.info("GCSClient.get_node({})".format(node_id))
        path = f"/nodes/{node_id}"
        return UnpackingGCSResponse(
            self.get(path, query_params=params, headers=self.headers), "node"
        )

    def update_node(
        self, node_id: Union[uuid.UUID, str], data: Dict, **params: Any
    ) -> GlobusHTTPResponse:
        """
        PATCH /nodes/{node_id}

        ** Parameters **

        ``node_id`` (*str*)
          UUID for the Node to be updated

        ``data`` (*dict*)
          Data in the format of a Node document.
        """
        log.info("GCSClient.update_node({}, ...)".format(node_id))
        path = f"/nodes/{node_id}"
        return self.patch(path, data=data, query_params=params, headers=self.headers)

    def delete_node(
        self, node_id: Union[uuid.UUID, str], **params: Any
    ) -> GlobusHTTPResponse:
        """
        DELETE /nodes/{node_id}

        ** Parameters **

        ``node_id`` (*str*)
          UUID for the Node to delete
        """
        log.info("GCSClient.delete_node({})".format(node_id))
        path = f"/nodes/{node_id}"
        return self.delete(path, query_params=params, headers=self.headers)

    #
    # Roles
    #

    @paging.has_paginator(
        paging.MarkerPaginator,
        items_key="data",
    )
    def get_role_list(
        self,
        collection_id: Union[uuid.UUID, str, None] = None,
        include: str = None,
        **params: Any,
    ) -> IterableGCSResponse:
        """
        GET /roles

        ** Paramaters **

        ``collection_id`` (*str*)
        UUID of a Collection. If given then only roles related to that
        Collection are returned, otherwise only Endpoint roles are returned.

        ``include`` (*str*)
        Pass "all_roles" to request all roles all roles relevant to the
        resource instead of only those the caller has on the resource
        """
        if include is not None:
            params = {**params, "include": include}
        if collection_id is not None:
            params = {**params, "collection_id": collection_id}

        path = "/roles"
        return IterableGCSResponse(
            self.get(path, query_params=params, headers=self.headers)
        )

    def create_role(self, data: Dict, **params: Any) -> UnpackingGCSResponse:
        """
        POST /roles

        ** Parameters **

        ``data`` (*dict*)
        Data in the format of a Role document, it is recommended to use
        the `RoleDocumment` class to construct this data.
        """
        path = "/roles"
        return UnpackingGCSResponse(
            self.post(path, data=data, query_params=params, headers=self.headers),
            "role",
        )

    def get_role(
        self, role_id: Union[uuid.UUID, str], **params: Any
    ) -> UnpackingGCSResponse:
        """
        GET /roles/{role_id}

        ** Parameters **

        ``role_id`` (*str*)
        UUID for the Role to be gotten
        """
        path = f"/roles/{role_id}"
        return UnpackingGCSResponse(
            self.get(path, query_params=params, headers=self.headers), "role"
        )

    def delete_role(
        self, role_id: Union[uuid.UUID, str], **params: Any
    ) -> GlobusHTTPResponse:
        """
        DELETE /roles/{role_id}

        ** Parameters **

        ``role_id`` (*str*)
        UUID for the Role to be deleted
        """
        path = f"/roles/{role_id}"
        return self.delete(path, query_params=params, headers=self.headers)

    #
    # Sharing Policy
    #

    def create_sharing_policy(self, data: Dict, **params: Any) -> UnpackingGCSResponse:
        """
        POST /sharing_policies

        ** Parameters **

        ``data`` (*dict*)
        Data in the format of a SharingPolicy document, it is recommended to
        use the `SharingPolicy` class to construct this data.
        """
        path = "/sharing_policies"
        return UnpackingGCSResponse(
            self.post(path, data=data, query_params=params, headers=self.headers),
            "sharing_policy",
        )

    @paging.has_paginator(
        paging.MarkerPaginator,
        items_key="data",
    )
    def get_sharing_policy_list(
        self, collection_id: Union[uuid.UUID, str], username: str = None, **params: Any
    ) -> IterableGCSResponse:
        """
        GET /sharing_policies

        ** Paramaters **

        ``collection_id`` (*str*)
        UUID of a Collection. Only policies related to this
        Collection are returned.

        ``username`` (*str*)
        Optional username. If present, only policies which apply to all users or
        to this user for the given collection are returned.
        """
        params = {**params, "collection_id": collection_id}
        if username is not None:
            params = {**params, "username": username}

        path = "/sharing_policies"
        return IterableGCSResponse(
            self.get(path, query_params=params, headers=self.headers)
        )

    def get_sharing_policy(
        self, sharing_policy_id: Union[uuid.UUID, str], **params: Any
    ) -> UnpackingGCSResponse:
        """
        GET /sharing_policies/{sharing_policy_id}

        ** Paramaters **

        ``sharing_policy_id`` (*str*)
        UUID of a Sharing Policy.
        """
        path = f"/sharing_policies/{sharing_policy_id}"
        return UnpackingGCSResponse(
            self.get(path, query_params=params, headers=self.headers), "sharing_policy"
        )

    def delete_sharing_policy(
        self, sharing_policy_id: Union[uuid.UUID, str], **params: Any
    ) -> GlobusHTTPResponse:
        """
        DELETE /sharing_policies/{sharing_policy_id}

        ** Parameters **

        ``sharing_policy_id`` (*str*)
        UUID for the SharingPolicy to delete
        """
        path = f"/sharing_policies/{sharing_policy_id}"
        return self.delete(path, query_params=params, headers=self.headers)

    #
    # User Credentials
    #

    @paging.has_paginator(
        paging.MarkerPaginator,
        items_key="data",
    )
    def get_user_credential_list(self, **params: Any) -> IterableGCSResponse:
        """
        GET /user_credentials
        """
        return IterableGCSResponse(
            self.get("/user_credentials", query_params=params, headers=self.headers)
        )

    def get_user_credential(
        self, user_credential_id: Union[uuid.UUID, str], **params: Any
    ) -> UnpackingGCSResponse:
        """
        GET /user_credentials/{user_credential_id}

        :param credential_id: The ID of the credential to fetch
        """
        path = f"/user_credentials/{user_credential_id}"
        return UnpackingGCSResponse(
            self.get(path, query_params=params, headers=self.headers), "credential"
        )
