from typing import (
    ClassVar,
    List,
    Optional,
    Type,
    TYPE_CHECKING,
    Union,
)

from .base import (
    Check,
    CheckInvoker,
    MissingPropertyCheck,
    ObjectIsNotNoneCheck,
    PropertyCheck,
    RelationshipChecker,
)

if TYPE_CHECKING:
    from ..models import (  # noqa: F401
        Credential,
        GuestCollection,
        MigrationPlan,
        StorageGateway,
    )


class MigrationPlanCheck(CheckInvoker["MigrationPlan"]):
    checkers: ClassVar[List[Type[Check["MigrationPlan"]]]] = []
    object_name = "migration_plan"


@MigrationPlanCheck.checker
class MP01(ObjectIsNotNoneCheck["MigrationPlan"]):
    ...


@MigrationPlanCheck.property_checker(property_name="storage_gateway")
class MP02(MissingPropertyCheck["MigrationPlan"]):
    ...


@MigrationPlanCheck.property_checker(property_name="mapped_collection")
class MP03(MissingPropertyCheck["MigrationPlan"]):
    ...


@MigrationPlanCheck.checker
class MP04(RelationshipChecker[Union["Credential", "StorageGateway"], "MigrationPlan"]):
    relationship_object_names = ["user_credential", "storage_gateway"]
    constraint_name = "identities must be a domain allowed by the storage gateway"

    def relationship_is_ok(
        self, obj: "MigrationPlan"
    ) -> List[List[Union["Credential", "StorageGateway"]]]:
        violations: List[List[Union["Credential", "StorageGateway"]]] = []

        sg = obj.storage_gateway
        if not sg:
            return violations
        if not sg.allowed_domains:
            return violations
        gc_list = obj.guest_collections
        if not gc_list:
            return violations

        for gc in gc_list:
            uc = gc.credential
            if uc:
                if uc.identity_username:
                    if uc.identity_username.rsplit("@", 1)[1] not in sg.allowed_domains:
                        violations.append([uc, sg])

        return violations


@MigrationPlanCheck.checker
class MP05(PropertyCheck["MigrationPlan"]):
    def value_is_ok(self, obj: "MigrationPlan") -> bool:
        if obj.guest_collections:
            if False in [g.disable for g in obj.guest_collections]:
                self.violations.append(obj)

        return bool(len(self.violations) == 0)

    @property
    def diagnostic(self) -> Optional[str]:
        if self.violations:
            return (
                "This version of the migration tools does not support shared endpoints."
            )
        return None
