import enum
from typing import (
    cast,
    Any,
    Generator,
    Iterable,
    List,
)

import click

import gcs_migration.check as checks
import gcs_migration.models as models
from gcs_migration.models import MigrationPlan
from .decorators import requires_logged_in, requires_migration_plan


class ObjectTypeChoices(enum.Enum):
    acl = "acl"
    credential = "credential"
    guest_collection = "guest-collection"
    mapped_collection = "mapped-collection"
    migration_plan = "migration-plan"
    role = "role"
    storage_gateway = "storage-gateway"


def get_acls(m: MigrationPlan) -> Generator[models.ACL, None, None]:
    if m.guest_collections:
        for g in m.guest_collections:
            if g.acls and not g.disable:
                for a in g.acls:
                    yield a


def get_credentials(m: MigrationPlan,) -> Generator[models.Credential, None, None]:
    if m.guest_collections:
        for g in m.guest_collections:
            if g.credential and not g.disable:
                yield g.credential


def get_guest_collections(
    m: MigrationPlan,
) -> Generator[models.GuestCollection, None, None]:
    if m.guest_collections:
        for g in m.guest_collections:
            if not g.disable:
                yield g


def get_roles(m: MigrationPlan) -> Generator[models.Role, None, None]:
    if m.mapped_collection:
        for mc in [m.mapped_collection]:
            for role in mc.roles or []:
                yield role
    if m.guest_collections:
        for g in m.guest_collections:
            if g.roles and not g.disable:
                for role in g.roles:
                    yield role


@click.command(
    "check",
    short_help="Check migration plan for consistency",
    help="""\
Check the migration plan for consistency, verifying that all identity mappings
are covered by the configuration, and that all configuration settings are
covered by this tool.

If not given any options, all objects are checked. Otherwise, only the objects
of the specified types are checked.
""",
)
@click.option(
    "--type-check",
    multiple=True,
    type=click.Choice([e.value for e in ObjectTypeChoices], case_sensitive=False),
)
@requires_logged_in
@requires_migration_plan
def check(type_check: Iterable[str], migration_plan: MigrationPlan) -> None:
    object_checkers = {
        ObjectTypeChoices.acl: (checks.ACLCheck, get_acls),
        ObjectTypeChoices.credential: (checks.CredentialCheck, get_credentials),
        ObjectTypeChoices.guest_collection: (
            checks.GuestCollectionCheck,
            get_guest_collections,
        ),
        ObjectTypeChoices.mapped_collection: (
            checks.MappedCollectionCheck,
            lambda x: [x.mapped_collection],
        ),
        ObjectTypeChoices.migration_plan: (checks.MigrationPlanCheck, lambda x: [x]),
        ObjectTypeChoices.role: (checks.RoleCheck, get_roles),
        ObjectTypeChoices.storage_gateway: (
            checks.StorageGatewayCheck,
            lambda x: [x.storage_gateway],
        ),
    }

    rc = 0

    if not type_check:
        type_check = [e.value for e in ObjectTypeChoices]

    for tc in type_check:
        check_name = ObjectTypeChoices[tc.replace("-", "_")]
        click.echo(f"Checking {check_name.name}...")
        checker_class, fetcher = object_checkers[check_name]

        checker = checker_class()
        violations: List[checks.Check[Any]] = []

        for datum in cast(Iterable, fetcher(migration_plan)):
            res = checker.check(datum)
            if res:
                rc += len(res)
                for r in res:
                    if r not in violations:
                        violations.append(r)
        for v in violations:
            print(
                f"{v.name}: {v.diagnostic if v.diagnostic else 'unable to execute check'}"
            )

    click.get_current_context().exit(rc)
