import json
import os
import pathlib
import shutil
from typing import Optional
from zipfile import BadZipFile, ZipFile

import click

from ..version import __version__


@click.command("restore")
@click.version_option(version=__version__)
@click.argument("SNAPSHOT_ID", type=int, required=False)
def snapshot_restore(snapshot_id: Optional[int]) -> None:
    migration_plan_dir = pathlib.Path("migration_plan")
    snapshot_dir = pathlib.Path("migration_snapshots")

    if snapshot_id is None and snapshot_dir.exists():
        snapshot_id = 0
        for fn in os.listdir(snapshot_dir):
            try:
                idx = int(fn)

                if idx > snapshot_id:
                    snapshot_id = idx
            except ValueError:
                pass

    if not snapshot_id:
        raise click.ClickException("No snapshots to restore")

    snapshot_path = snapshot_dir / f"{snapshot_id:04}"
    if not snapshot_path.exists():
        raise click.ClickException(f"Snapshot {snapshot_id} does not exist.")

    try:
        with ZipFile(snapshot_path, mode="r") as z:
            if z.testzip() is not None:
                raise click.ClickException("Invalid snapshot")
            if migration_plan_dir.exists():
                shutil.rmtree("migration_plan")

            with z.open("metadata.json", "r") as f:
                metadata = json.load(f)

            namelist = [n for n in z.namelist() if n != "metadata.json"]

            def showit(item: Optional[str]) -> str:
                if item:
                    return item.replace("migration_plan/", "")
                else:
                    return ""

            with click.progressbar(
                namelist, item_show_func=showit, show_eta=False
            ) as bar:
                for item in bar:
                    z.extract(item)
    except BadZipFile:
        raise click.ClickException("Unable to parse snapshot")

    if "description" in metadata:
        click.echo(f"Restored to snapshot {snapshot_id}: {metadata['description']}")
    else:
        click.echo(f"Restored to snapshot {snapshot_id}")
