import os
import pathlib
import pwd
import re
from typing import List, Optional, TYPE_CHECKING, Union, cast
from urllib.parse import unquote

from .acl import ACL
from .credential import Credential
from .endpoint_metadata import (
    EndpointMetadata,
    EndpointMetadataDict,
)
from ..exceptions import UnsupportedUpdate
from .role import Role, RoleDict
from ..util import yaml, yaml_object
from ..exceptions import InvalidGuestCollectionException

if TYPE_CHECKING:
    from .credential import CredentialDict  # noqa: F401
    from ..typing.transfer_documents import (  # noqa: F401
        AccessListDocumentDict,
        AnonymousEffectivePermissionsDict,
        EndpointDocumentDict,
    )
    from ..typing.gcs_documents import (  # noqa: F401
        PosixGuestCollectionDict,
        PosixCollectionPoliciesDict,
    )


class GuestCollectionDict(EndpointMetadataDict):
    host_path: Optional[str]
    base_path: Optional[str]
    default_directory: Optional[str]
    force_encryption: Optional[bool]
    disable_verify: Optional[bool]
    force_verify: Optional[bool]
    share_state_file: Optional[str]
    owner_id: Optional[str]
    owner_string: Optional[str]
    enable_https: Optional[bool]
    mapped_collection_id: Optional[str]
    disable: bool


@yaml_object(yaml)
class GuestCollection(EndpointMetadata):
    def __init__(self, data: Optional[GuestCollectionDict] = None) -> None:
        super().__init__()

        self.yaml_path: Optional[str] = None
        self.id: Optional[str] = None
        self.host_path: Optional[str] = None
        self.base_path: Optional[str] = None
        self.default_directory: Optional[str] = None
        self.force_encryption: Optional[bool] = None
        self.disable_verify: Optional[bool] = None
        self.force_verify: Optional[bool] = None
        self.owner_id: Optional[str] = None
        self.owner_string: Optional[str] = None
        self.acls: Optional[List[ACL]] = None
        self.roles: Optional[List[Role]] = None
        self.credential: Optional[Credential] = None
        self.user_cert_chain: Optional[str] = None
        self.sharing_uuid: Optional[str] = None
        self.enable_https: Optional[bool] = None
        self.mapped_collection_id: Optional[str] = None
        self.transfer_endpoint: Optional["EndpointDocumentDict"] = None
        self.access_list: Optional[AccessListDocumentDict] = None
        self.anonymous_effective_permissions: Optional[
            "AnonymousEffectivePermissionsDict"
        ] = None
        self.share_state_file: Optional[Union[str, os.PathLike]] = None
        self.disable = False
        if data:
            self.__setstate__(data)

    @property
    def transfer_endpoint(self) -> Optional["EndpointDocumentDict"]:
        return self._transfer_endpoint

    @transfer_endpoint.setter
    def transfer_endpoint(self, value: Optional["EndpointDocumentDict"]) -> None:
        self._transfer_endpoint = value
        if value is not None:
            if self.default_directory is None:
                self.default_directory = value["default_directory"]
            if self.force_encryption is None:
                self.force_encryption = value["force_encryption"]
            if self.disable_verify is None:
                self.disable_verify = value["disable_verify"]
            if self.force_verify is None:
                self.force_verify = value["force_verify"]
            if self.owner_id is None:
                self.owner_id = value["owner_id"]
            if self.owner_string is None:
                self.owner_string = value["owner_string"]
            if self.host_path is None:
                self.host_path = value["host_path"]

    @property
    def access_list(self) -> Optional["AccessListDocumentDict"]:
        return self._access_list

    @access_list.setter
    def access_list(self, value: Optional["AccessListDocumentDict"]) -> None:
        self._access_list = value
        if value is not None:
            roles = []
            acls = []
            self._access_list = value
            for access in value.get("DATA", []):
                if access.get("role_id"):
                    role = Role()
                    role.access_document = access
                    roles.append(role)
                else:
                    acl = ACL()
                    acl.access_document = access
                    acl.collection_id = self.id
                    acls.append(acl)
            self.roles = roles
            self.acls = acls

    @property
    def acls(self) -> Optional[List[ACL]]:
        return self._acls

    @acls.setter
    def acls(self, value: Optional[List[ACL]]) -> None:
        self._acls = value

    @property
    def roles(self) -> Optional[List[Role]]:
        return self._roles

    @roles.setter
    def roles(self, value: Optional[List[Role]]) -> None:
        self._roles = value

    @property
    def anonymous_effective_permissions(
        self,
    ) -> Optional["AnonymousEffectivePermissionsDict"]:
        return self._anonymous_effective_permissions

    @anonymous_effective_permissions.setter
    def anonymous_effective_permissions(
        self, value: Optional["AnonymousEffectivePermissionsDict"]
    ) -> None:
        self._anonymous_effective_permissions = value
        if value is not None:
            if self.user_cert_chain is None:
                self.user_cert_chain = value["user_cert_chain"]
            if self.sharing_uuid is None:
                self.sharing_uuid = value["sharing_uuid"]

    def add_roles_for_old_owner(self) -> None:
        if self.owner_id:
            if self.roles is None:
                self.roles = []
            if (
                self.credential
                and self.credential.identity_id
                and self.credential.identity_id == self.owner_id
            ):
                return

            for role in self.roles:
                if (
                    role.principal == self.owner_id
                    and role.principal_type == "identity"
                    and role.role == "administrator"
                ):
                    return

            role_data: RoleDict = {
                "id": None,
                "principal": self.owner_id,
                "principal_type": "identity",
                "role": "administrator",
            }

            self.roles.append(Role(data=role_data))

    def load_share_state_file(self) -> None:
        path = self.share_state_file
        if not path:
            return

        share_file = path if isinstance(path, str) else path.__fspath__()
        need_home = False
        need_user = False

        if "~" in share_file or "$HOME" in share_file:
            need_home = True
        if "$USER" in share_file:
            need_user = True

        if need_home or need_user:
            if not self.credential or not self.credential.username:
                return

            home = None
            if need_home:
                try:
                    pwent = pwd.getpwnam(self.credential.username)
                    home = pwent.pw_dir
                    if not home:
                        return
                except KeyError:
                    return

            assert home is not None
            share_file = share_file.replace("$USER", self.credential.username, 1)
            share_file = share_file.replace("$HOME", home, 1)
            share_file = share_file.replace("~", home, 1)

        try:
            with open(share_file, "r") as f:
                share_path = None
                for line in f:
                    line = re.sub("#.*", "", line)
                    line = re.sub(r"^\s*", "", line)
                    if not line:
                        continue
                    m = re.match(r'share_path\s+"(?P<share_path>[^"]*)"', line)
                    if m:
                        share_path = unquote(m.group("share_path"))
                        break
                self.base_path = share_path
        except Exception:
            # Ignore I/O errors. The share_state_file property will be
            # set, so we can try again later
            return

    @property
    def share_state_file(self) -> Optional[Union[str, os.PathLike]]:
        return self._share_state_file

    @share_state_file.setter
    def share_state_file(self, value: Optional[Union[str, os.PathLike]]) -> None:
        self._share_state_file = value

        if self.base_path is None and value:
            self.load_share_state_file()

    @property
    def base_path(self) -> Optional[str]:
        return self._base_path

    @base_path.setter
    def base_path(self, value: Optional[str]) -> None:
        self._base_path = value

    @property
    def credential(self) -> Optional[Credential]:
        return self._credential

    @credential.setter
    def credential(self, value: Optional[Credential]) -> None:
        self._credential = value

        if value and self.share_state_file and not self.base_path:
            self.load_share_state_file()

    def dump(self, root: "Union[str, os.PathLike[str]]") -> None:
        assert self.id is not None

        path = pathlib.Path(root) / self.id
        if not os.path.exists(path):
            os.makedirs(path, mode=0o755)

        with open(path / "collection.yaml", "w") as f:
            yaml.dump(self, f)

        if self.acls:
            with open(path / "acls.yaml", "w") as f:
                yaml.dump(self.acls, f)

        if self.roles:
            with open(path / "roles.yaml", "w") as f:
                yaml.dump(self.roles, f)

        with open(path / "credential.yaml", "w") as f:
            yaml.dump(self.credential, f)

    @staticmethod
    def load(
        root: "Union[str, os.PathLike[str]]", id: str
    ) -> Optional["GuestCollection"]:
        path = pathlib.Path(root) / id

        collection = path / "collection.yaml"

        if not os.path.exists(collection):
            return None

        with open(collection, "r") as f:
            obj = yaml.load(f)
            obj.id = id
            obj.yaml_path = str(collection)
            assert isinstance(obj, GuestCollection)

        acls = path / "acls.yaml"
        acl_objs: List[ACL] = []
        if os.path.exists(acls):
            with open(acls, "r") as f:
                acl_objs = yaml.load(f)
                assert isinstance(acl_objs, list)
                for acl_obj in acl_objs:
                    assert isinstance(acl_obj, ACL)
                    acl_obj.collection_id = obj.id
                    acl_obj.yaml_path = str(acls)
        obj.acls = acl_objs

        roles = path / "roles.yaml"
        role_objs: List[Role] = []
        if os.path.exists(roles):
            with open(roles, "r") as f:
                role_objs = yaml.load(f)
                assert isinstance(role_objs, list)
                for role_obj in role_objs:
                    assert isinstance(role_obj, Role)
        obj.roles = role_objs

        credential = path / "credential.yaml"
        if os.path.exists(credential):
            with open(credential, "r") as f:
                credential_obj = yaml.load(f)
                assert isinstance(credential_obj, Credential)
                credential_obj.collection_id = obj.id
                obj.credential = credential_obj
        else:
            obj.credential = None

        return obj

    def update(self, other: "GuestCollection") -> None:  # type: ignore
        for attr in [
            "host_path",
            "base_path",
            "default_directory",
            "force_encryption",
            "disable_verify",
            "force_verify",
            "acls",
            "roles",
            "enable_https",
            "mapped_collection_id",
        ]:
            old_value = getattr(self, attr, None)
            new_value = getattr(other, attr, None)
            if old_value != new_value:
                setattr(self, attr, new_value)

        if self.owner_string != other.owner_string:
            self.owner_string = other.owner_string
            self.credential = other.credential

        unsupported_updates = []
        for attr in [
            "owner_id",
            "share_state_file",
        ]:
            old_value = getattr(self, attr, None)
            new_value = getattr(other, attr, None)
            if new_value and str(old_value) != str(new_value):
                unsupported_updates.append(
                    f"Unable to update {attr} from {old_value} to {new_value}"
                )

        if unsupported_updates:
            raise UnsupportedUpdate("\n".join(unsupported_updates))

        super().update(other)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, GuestCollection):
            return NotImplemented

        return self.__getstate__() == other.__getstate__()

    def __getstate__(self) -> GuestCollectionDict:
        metadata_state = super().__getstate__()

        share_state_file = self.share_state_file

        if isinstance(share_state_file, os.PathLike):
            share_state_file = share_state_file.__fspath__()

        return {
            "host_path": self.host_path,
            "base_path": self.base_path,
            "default_directory": self.default_directory,
            "force_encryption": self.force_encryption,
            "disable_verify": self.disable_verify,
            "force_verify": self.force_verify,
            "share_state_file": share_state_file,
            "owner_id": self.owner_id,
            "owner_string": self.owner_string,
            "enable_https": self.enable_https,
            "mapped_collection_id": self.mapped_collection_id,
            "disable": self.disable,
            **metadata_state,  # type: ignore
        }

    def __setstate__(self, state: GuestCollectionDict) -> None:  # type: ignore
        super().__setstate__(state)

        if not hasattr(self, "credential"):
            self.credential = None
        if not hasattr(self, "acls"):
            self.acls = None

        self.host_path = state["host_path"]
        self.base_path = state["base_path"]
        self.default_directory = state["default_directory"]
        self.force_encryption = state["force_encryption"]
        self.disable_verify = state["disable_verify"]
        self.force_verify = state["force_verify"]
        self.share_state_file = state["share_state_file"]
        self.owner_id = state["owner_id"]
        self.owner_string = state["owner_string"]
        self.enable_https = state["enable_https"]
        self.mapped_collection_id = state["mapped_collection_id"]
        self.disable = state["disable"]

    def to_json(self) -> "PosixGuestCollectionDict":
        if self.disable:
            raise InvalidGuestCollectionException(
                f"Guest collection {self.id} is disabled"
            )

        policies: "PosixCollectionPoliciesDict" = {
            "DATA_TYPE": "posix_collection_policies#1.1.0"
        }
        collection = cast("PosixGuestCollectionDict", super().to_json())
        collection["DATA_TYPE"] = "collection#1.4.0"
        collection["policies"] = policies
        collection["collection_type"] = "guest"
        if self.credential and self.credential.identity_id:
            collection["identity_id"] = self.credential.identity_id

        if self.mapped_collection_id:
            collection["mapped_collection_id"] = self.mapped_collection_id

        if self.base_path:
            collection["collection_base_path"] = self.base_path
        elif self.host_path:
            collection["collection_base_path"] = self.host_path

        if self.public is None:
            collection["public"] = True
        else:
            collection["public"] = self.public

        if self.default_directory:
            collection["default_directory"] = self.default_directory

        if self.force_encryption is not None:
            collection["force_encryption"] = self.force_encryption

        if self.disable_verify is not None:
            collection["disable_verify"] = self.disable_verify

        if self.enable_https is not None:
            collection["enable_https"] = self.enable_https

        if self.force_verify is not None:
            collection["force_verify"] = self.force_verify

        return collection

    def __str__(self) -> str:
        id_str = f" {self.id}" if self.id else ""
        display_name_str = f" {self.display_name}" if self.display_name else ""
        return f"<Guest Collection{id_str}{display_name_str}>"
