from typing import List, Optional
from typing_extensions import TypedDict

from cryptography import x509

from ..parsers import CalloutDict
from ..util import yaml, yaml_object


class MyProxyConfigurationDict(TypedDict):
    dn: Optional[str]
    callout: Optional[CalloutDict]
    domains: List[str]


@yaml_object(yaml)
class MyProxyConfiguration:
    def __init__(self, callout: Optional[CalloutDict] = None) -> None:
        self.dn: Optional[str] = None
        self.ca_cert: Optional[x509.Certificate] = None
        self.callout = callout
        self.domains: List[str] = []

    @property
    def is_myproxy(self) -> bool:
        return True

    @property
    def is_cilogon(self) -> bool:
        return False

    @property
    def is_gridmap(self) -> bool:
        return False

    @property
    def ca_cert(self) -> Optional[x509.Certificate]:
        return self._ca_cert

    @ca_cert.setter
    def ca_cert(self, value: Optional[x509.Certificate]) -> None:
        self._ca_cert = value

        if value and not self.dn:
            self.dn = "/" + "/".join(
                [rdn.rfc4514_string() for rdn in value.subject.rdns]
            )

    @property
    def callout(self) -> Optional[CalloutDict]:
        return self._callout

    @callout.setter
    def callout(self, value: Optional[CalloutDict]) -> None:
        self._callout = value

        if value and not self.dn:
            cert_path = value["environment"].get("GLOBUS_MYPROXY_CA_CERT")
            cert_data = None
            if cert_path:
                with open(cert_path, "rb") as f:
                    cert_data = f.read()
            if cert_data:
                self.ca_cert = x509.load_pem_x509_certificate(cert_data)

    def add_domain(self, domain: str) -> None:
        if domain not in self.domains:
            self.domains.append(domain)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MyProxyConfiguration):
            return NotImplemented

        return self is other or self.__getstate__() == other.__getstate__()

    def __getstate__(self) -> MyProxyConfigurationDict:
        return {"dn": self.dn, "callout": self.callout, "domains": self.domains}

    def __setstate__(self, state: MyProxyConfigurationDict) -> None:
        self.dn = state["dn"]
        self.callout = state["callout"]
        self.domains = state["domains"]
