from subprocess import Popen, PIPE
from typing import Dict, List, Optional

from flask import current_app
import marshmallow as ma
import globus.manager.api.errors as errors


class MappingInputSchema(ma.Schema):
    _DATA_TYPE = "identity_mapping_input#1.0.0"
    DATA_TYPE = ma.fields.Str(
        default=_DATA_TYPE, validate=[ma.validate.Equal(_DATA_TYPE)]
    )

    identities = ma.fields.List(ma.fields.Raw)


class MappingResultSchema(ma.Schema):
    id = ma.fields.UUID()
    output = ma.fields.Str()


class MappingOutputSchema(ma.Schema):
    _DATA_TYPE = "identity_mapping_output#1.0.0"
    DATA_TYPE = ma.fields.Str(
        default=_DATA_TYPE, validate=[ma.validate.Equal(_DATA_TYPE)]
    )
    result = ma.fields.List(ma.fields.Nested(MappingResultSchema))


identity_mapping_input_schema = MappingInputSchema()
identity_mapping_output_schema = MappingOutputSchema()


class ExternalIdentityMapping:
    """
    The ExternalIdentityMapping class implements an extension point for
    administrators to impement their own identity mapping algorithm as an
    external progam.

    The GCS Manager will run the configured program and arguments in a sub
    process. In addition to any arguments specified in the configuration
    document, the following command-line options will be added:

    The `-c` option along with an argument which is the UUID of a
    GCS connector.

    The `-s` option along with its argument which is the UUID of the
    storage gateway mapping the identity.

    The `-a` command-line option indicates that the GCS Manager wants to
    receive output containing all mappings for the given identity set. If
    present, the program will receive an input containing at least one identity
    resource as input, and may return as many mappings as can be computed from
    the identities. If not present, the program will receive exactly one object
    in the identities orray and may only return a single mapping for that
    identity. If it returns multiple mappings, only the first is used by
    the GCS Manager.

    The GCS Manager communicates between the external program using
    JSON-formatted input and output data passed via stdin and stdout.

    Input consists of a JSON object containing two properties: `DATA_TYPE`
    which indicates the version of the input document, currently
    `identity_mapping_input#1.0.0` and `identities`, which contains a list of
    identity resource obtained from Globus Auth.

    For example:

    .. code-block:: json

        {
            "DATA_TYPE": "identity_mapping_input#1.0.0",
            "identities": [
                {
                    "id": "9cf5a1d2-41c4-4925-b0bb-f5e0a16c98ab",
                    "username": "user@example.com",
                    "identity_provider": "1f8ba017-6224-43ad-a42f-03059574ba38"
                },
                {
                    "id": "67c31064-1e5f-4d78-b687-5fbea3d3e05a",
                    "username": "user2@example.org",
                    "identity_provider": "05f6f886-9277-42b9-836e-81559f337ac3"
                },
                {
                    "id": "4787931e-b999-412e-97f9-a7a24c79ad71",
                    "username": "user3@example.info",
                    "identity_provider": "05f6f886-9277-42b9-836e-81559f337ac3"
                }
            ]
        }

    Output is an object containing two properties, `DATA_TYPE` which
    indicates the version of the output document, currently
    `identity_mapping_output#1.0.0`, and `result`, which contains a list of
    objects with the properties `id` and `output` The `id` is the value of one
    of the `id` properties of one of the input identities, and the `output` is
    a connector-specific user identifier that indicates the result of the
    mapping.

    For example:

    .. code-block:: json

        {
            "DATA_TYPE": "identity_mapping_output#1.0.0",
            "result": [
                {
                    "id": "67c31064-1e5f-4d78-b687-5fbea3d3e05a",
                    "output": "user3"
                },
            ]
        }
    """

    DATA_TYPE = "external_identity_mapping#1.0.0"
    mapping_timeout = 5

    def __init__(self, *args: str, storage_gateway: str = None, connector: str = None):
        """
        :param `*args`:
            List of strings which are the path to the external command and the
            command line arguments to pass to the module. Additional paramters
            `-s`, `-c`, and `-a` will be automatically added to this
            command-line as described in the class documentation.

        :param storage_gateway:
            String contaiing the UUID of the storage gateway to perform the
            mapping for.
        :param connector: String contaiing the UUID of the connector
            to perform the mapping for.
        """
        assert len(args) > 0
        assert connector is not None
        assert storage_gateway is not None

        self.command = list(args)
        self.connector = connector
        self.storage_gateway = storage_gateway

    @staticmethod
    def from_json(mapping_document, connector: str = None, storage_gateway: str = None):
        if mapping_document["DATA_TYPE"] != ExternalIdentityMapping.DATA_TYPE:
            raise ValueError("Unsupported DATA_TYPE")

        return ExternalIdentityMapping(
            *mapping_document["command"],
            connector=connector,
            storage_gateway=storage_gateway
        )

    def map_identity(self, identity_data: Dict[str, str]) -> Optional[str]:
        """
        Map a single identity object to a username for this particular
        connector and storage gateway.

        :param identity:
            A dict containing the identity assertions from Globus Auth.

        :raises errors.InternalServerError:
            Error executing the external mapping program, or the output
            from the program is not in an understood format.
        """
        # TODO: set a timeout?
        p = Popen(
            self.command + ["-c", self.connector, "-s", self.storage_gateway],
            universal_newlines=True,
            stdout=PIPE,
            stderr=PIPE,
            stdin=PIPE,
        )

        input_data = identity_mapping_input_schema.dumps(
            {"identities": [identity_data]}
        )

        stdout, stderr = p.communicate(input_data, timeout=self.mapping_timeout)

        if p.returncode != 0:
            current_app.logger.debug("Mapping error: stdout=%s", stdout)
            current_app.logger.debug("Mapping error: stderr=%s", stderr)
            raise errors.InternalServerError("Error mapping identity")

        if stdout:
            try:
                output = identity_mapping_output_schema.loads(stdout)
            except Exception as e:
                current_app.logger.error(
                    "The mapping application %s yielded an invalid result %s, "
                    "causing %s",
                    self.command,
                    stdout,
                    str(e),
                )
                raise errors.InternalServerError()
            if output["result"]:
                first_result = output["result"][0]
                if str(first_result["id"]) != identity_data["id"]:
                    raise errors.InternalServerError(
                        "Invalid response from mapping program"
                    )

                return first_result["output"]

        return None

    def map_identities(
        self, identity_data: List[Dict[str, str]]
    ) -> List[Dict[str, List[str]]]:
        """
        Map all identities from the associated input list to local usernames
        that can potentially access the configured Storage Gateway

        :param identity_data: List of identity assertions from Globus Auth

        :returns:
            List of mappings from the `id` property of the input data
            to a list of local username strings that the identity maps to.

        :raises errors.InternalServerError:
            Error executing the external mapping program, or the output
            from the program is not in an understood format.
        """
        results = []  # type: List[Dict[str, List[str]]]

        p = Popen(
            self.command + ["-c", self.connector, "-a"],
            universal_newlines=True,
            stdout=PIPE,
            stderr=PIPE,
            stdin=PIPE,
        )
        input_data = identity_mapping_input_schema.dumps({"identities": identity_data})
        stdout, stderr = p.communicate(input_data, timeout=self.mapping_timeout)
        if p.returncode != 0:
            raise errors.InternalServerError(
                "Error running mapping program {}".format(self.command[0])
            )
        if stdout:
            input_ids = [i["id"] for i in identity_data]
            try:
                output = identity_mapping_output_schema.loads(stdout)
            except Exception as e:
                current_app.logger.error(
                    "The mapping application %s yielded an invalid result %s, "
                    "causing %s",
                    self.command,
                    stdout,
                    str(e),
                )
                raise errors.InternalServerError()

            result = {}  # type: Dict[str, List[str]]

            for o in output["result"]:
                output_id = str(o["id"])
                if output_id not in input_ids:
                    raise errors.InternalServerError("Invalid ID in result")
                if output_id not in result:
                    result[output_id] = []
                result[output_id].append(o["output"])
            results.append(result)
        return results
