import ipaddress
import uuid

from globus.manager.constants import NodeStatus


class Node:
    """
    Model for a GCS Node.

    Each Node object in the API consists of:
      - [required] an ID of type UUID
      - [required] one or more IP addresses
      - [optional] a status field of type NodeStatus, defaults to active
      - [optional] incoming port range
      - [optional] outgoing port range
      - [optional] data interface
    """

    _DATA_TYPE = "node#1.1.0"
    _DATA_TYPES = ["node#1.0.0", "node#1.1.0"]

    def __init__(
        self,
        DATA_TYPE=None,
        ip_addresses=None,
        id=None,
        status=None,
        incoming_port_range=None,
        outgoing_port_range=None,
        data_interface=None,
    ):
        if not DATA_TYPE or not DATA_TYPE.startswith("node#1."):
            raise ValueError("Invalid DATA_TYPE for node")

        if id is None:
            self._id = str(uuid.uuid4())
        else:
            self._id = id

        self.ip_addresses = ip_addresses
        if not status:
            self.status = NodeStatus.active
        elif isinstance(status, NodeStatus):
            self.status = status
        else:
            try:
                self.status = getattr(NodeStatus, status)
            except AttributeError:
                raise ValueError("Invalid node status {}".format(status))
        if self.ip_addresses is None:
            raise ValueError("The ip_addresses property for a node may not be empty")

        self.incoming_port_range = incoming_port_range
        self.outgoing_port_range = outgoing_port_range
        self.data_interface = data_interface

    @property
    def id(self):
        return self._id

    @property
    def ip_addresses(self):
        return self._ip_addresses

    @ip_addresses.setter
    def ip_addresses(self, value):
        if value is None or len(value) == 0:
            raise ValueError("Invalid ip_addresses")
        self._ip_addresses = [ipaddress.ip_address(ip) for ip in value]

    @property
    def incoming_port_range(self):
        if not self._incoming_port_range:
            raise AttributeError(
                "The incoming_port_range property is not defined for this node"
            )
        return self._incoming_port_range

    @staticmethod
    def _is_valid_port_range(start, end):
        return (
            start == int(start)
            and end == int(end)
            and start < end
            and start > 1
            and start < 2 ** 16 - 1
            and end > 1
            and end < 2 ** 16 - 1
        )

    @incoming_port_range.setter  # type: ignore
    def incoming_port_range(self, value):
        if (
            isinstance(value, (tuple, list))
            and len(value) == 2
            and self._is_valid_port_range(value[0], value[1])
        ):
            self._incoming_port_range = tuple(value)
        elif not value:
            self._incoming_port_range = None
        else:
            raise ValueError("Invalid incoming_port_range")

    @property
    def outgoing_port_range(self):
        if not self._outgoing_port_range:
            raise AttributeError(
                "The outgoing_port_range property is not defined for this node"
            )
        return self._outgoing_port_range

    @outgoing_port_range.setter  # type: ignore
    def outgoing_port_range(self, value):
        if (
            isinstance(value, (tuple, list))
            and len(value) == 2
            and self._is_valid_port_range(value[0], value[1])
        ):
            self._outgoing_port_range = tuple(value)
        elif not value:
            self._outgoing_port_range = None
        else:
            raise ValueError("Invalid outgoing_port_range")

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, value):
        self._status = NodeStatus(value)

    @property
    def data_interface(self):
        if not self._data_interface:
            raise AttributeError(
                "The data_interface property is not defined for this node"
            )
        return self._data_interface

    @data_interface.setter
    def data_interface(self, value):
        self._data_interface = None
        if value is not None:
            self._data_interface = ipaddress.ip_address(value)

    def put(self, node):
        changed = False
        if not isinstance(node, Node):
            if node.get("DATA_TYPE") not in Node._DATA_TYPES:
                raise ValueError("Unknown DATA_TYPE")

            node_id_str = str(node["id"])
            ip_addresses = sorted(
                [ipaddress.ip_address(ip) for ip in node["ip_addresses"]]
            )
            status = NodeStatus(node["status"])
            incoming_port_range = node.get("incoming_port_range")
            outgoing_port_range = node.get("outgoing_port_range")
            data_interface = node.get("data_interface")
        else:
            node_id_str = str(node.id)
            ip_addresses = sorted(node.ip_addresses)
            status = NodeStatus(node.status)
            incoming_port_range = getattr(node, "incoming_port_range", None)
            outgoing_port_range = getattr(node, "outgoing_port_range", None)
            data_interface = getattr(node, "data_interface", None)

        if node_id_str != self._id:
            raise ValueError("The id property is immutable")

        if ip_addresses != self._ip_addresses:
            self.ip_addresses = ip_addresses
            changed = True

        if status != self._status:
            self._status = status
            changed = True

        if incoming_port_range != self._incoming_port_range:
            self.incoming_port_range = incoming_port_range
            changed = True

        if outgoing_port_range != self._outgoing_port_range:
            self.outgoing_port_range = outgoing_port_range
            changed = True

        if data_interface != self._data_interface:
            self.data_interface = data_interface
            changed = True

        return changed

    def patch(self, node_changes):
        changed = False
        for n in node_changes:
            if n == "DATA_TYPE":
                if node_changes[n] not in Node._DATA_TYPES:
                    raise ValueError("Unknown DATA_TYPE")
            elif n == "id":
                if str(node_changes[n]) != self.id:
                    raise ValueError("id is an immutable property")
            else:
                old_value = getattr(self, n, None)
                new_value = node_changes[n]
                if old_value != new_value:
                    setattr(self, n, new_value)
                    changed = True
        return changed

    def __eq__(self, other):
        return (
            self._id == other._id
            and self._ip_addresses == other._ip_addresses
            and self._status == other._status
            and self._incoming_port_range == other._incoming_port_range
            and self._outgoing_port_range == other._outgoing_port_range
            and self._data_interface == other._data_interface
        )

    def __repr__(self):
        return "<Node {}>".format(self.id)
