from typing import Union, Optional, TYPE_CHECKING
import uuid

from flask import current_app
from sqlalchemy.ext.hybrid import hybrid_property
import globus_sdk

from .decorators import DocumentChecker, checks_property
from .endpoint import Endpoint
from . import db, ConfigurationObject
from sqlalchemy.sql import func
from globus.manager.constants import (
    ConfigurationObjectTag,
    RoleType,
    PrincipalType,
)

import globus.manager.api.errors as errors
from globus.manager.api.schemas.appsync.role_config import role_config_schema
from globus.manager.api.schemas.transfer.role import role_schema
from globus.manager.api.utils.log import log_transfer_exception


if TYPE_CHECKING:
    from . import Collection


class Role(ConfigurationObject, DocumentChecker):
    """
    This is the model of the Role type, used by the role
    authorization code. In general, roles can not be altered after being
    created. The constraints prevent creation of multiple roles
    that grant the same permissions to the same principal for the same
    collection or the endpoint itself.
    """

    __tablename__ = "role"
    __table_args__ = (
        # We can ignore this uniqueness constraint when creating new role
        # objects, as transfer is the one issuing the IDs and implements
        # its own version of this
        db.UniqueConstraint(
            "principal_type",
            "identity_id",
            "role",
            "resource",
            name="unique_collection_role",
        ),
    )
    __mapper_args__ = {"polymorphic_identity": ConfigurationObjectTag.ROLE.name}

    configuration_schema = role_config_schema

    # Database columns

    # ForeignKey connecting this object to the related configuration_object
    # This ID is assigned in the initial object constructor based on the
    # response from transfer when creating the role
    id = db.Column(db.TEXT, db.ForeignKey("configuration_object.id"), primary_key=True)

    # Discriminate between identity and group interpretation of the identity_id
    # column
    principal_type = db.Column("principal_type", db.Enum(PrincipalType), nullable=False)

    # Identity to which this role is assigned
    _identity_id = db.Column("identity_id", db.TEXT, nullable=False)

    # Scoped role name
    role = db.Column("role", db.Enum(RoleType), nullable=False)

    # Resource to which the role grants permissions
    _resource = db.Column(
        "resource", db.TEXT, db.ForeignKey("collection.id"), nullable=True
    )

    primary_role = db.Column(db.BOOLEAN, nullable=False, server_default="0")

    def __init__(
        self,
        appsync_data=None,
        doc=None,
        primary_role=False,
        endpoint: Optional[Endpoint] = None,
    ):
        if appsync_data is not None:
            # Loading from appsync
            super().__init__(**{"tag": ConfigurationObjectTag.ROLE, **appsync_data})
            self.init_appsync()
        elif doc.get("DATA_TYPE") == "role#1.0.0":
            # Loading from the API schema which describes roles differently
            # than how it is stored in the database:
            # - The scope of the (role endpoint: or collection:) is not
            #   included in the role name, but is implicit in whether the role
            #   includes a collection property.
            # - The principal property uses a URN-based format instead of
            #   splitting the principal_type and identity_id into separate
            #   fields
            self.check_doc(**doc)
            collection = doc.get("collection", None)
            principal = doc.get("principal")
            role = doc.get("role")

            self.principal = principal  # type: ignore

            if role == "owner" and self.principal_type != PrincipalType.identity:
                raise ValueError("The owner role may only be assigned to an identity")

            if role == "owner" and collection:
                raise ValueError("Owner role may not be created for collections")

            if collection:
                resource = str(collection)
                self.role = RoleType("collection:" + role).name
                endpoint_id = resource
                self.resource = resource  # type: ignore
            else:
                self.role = RoleType("endpoint:" + role).name
                assert endpoint is not None
                endpoint_id = endpoint.endpoint_id

            role_id = self._create_role(endpoint_id)

            super().__init__(tag=ConfigurationObjectTag.ROLE, id=role_id)
            self.primary_role = primary_role
            self.refresh_configuration_hash()
        else:
            raise ValueError("Unknown DATA_TYPE")

        if not self.id:
            raise ValueError("Missing id")
        if not self.principal_type:
            raise ValueError("Missing principal_type")
        if not self.identity_id:
            raise ValueError("Missing identity_id")

    def init_appsync(self):
        ch = self.configuration_hash
        self.id = ch["id"]
        self.principal_type = ch["principal_type"]
        self._identity_id = ch["identity_id"]
        self.role = ch["role"]
        self.resource = ch.get("resource")
        self.primary_role = ch.get("primary_role", False)

    @db.validates("principal_type")
    def principal_type_immutable(self, key, value):
        if self.principal_type:
            old_pt = self.principal_type
        else:
            old_pt = None
        if value:
            pt = PrincipalType(value)
        else:
            pt = None
        if old_pt and old_pt != pt:
            raise ValueError("The principal_type property is immutable")
        return pt

    @hybrid_property
    def identity_id(self):
        return self._identity_id

    @identity_id.expression  # type: ignore
    def identity_id(cls):
        return cls._identity_id

    @identity_id.setter  # type: ignore
    def identity_id(self, value: Union[str, uuid.UUID]):
        if isinstance(value, uuid.UUID):
            identity_id_str = str(value)
        else:
            identity_id_str = value
        self._identity_id = identity_id_str

    @checks_property(prop="identity_id")
    def check_identity_id(self, value: Union[str, uuid.UUID]):
        if isinstance(value, uuid.UUID):
            identity_id_str = str(value)
        else:
            identity_id_str = value

        if self.identity_id and self.identity_id != identity_id_str:
            raise ValueError("The identity_id property is immutable")

    def put_owner_role(self, value):
        """
        The owner role is the only one which may be changed after creation, but
        only changing the identity ID that the role belongs to.
        """
        if RoleType(self.role) != RoleType.endpoint_owner:
            raise errors.InternalServerError()

        if isinstance(value, uuid.UUID):
            identity_id_str = str(value)
        else:
            try:
                identity_id_str = str(uuid.UUID(value))
            except ValueError:
                raise errors.InternalServerError()

        self._identity_id = identity_id_str
        self.refresh_configuration_hash()
        return self._identity_id

    @db.validates("role")
    def role_validate(self, key, value):
        if not isinstance(value, RoleType):
            if ":" in value:
                r = RoleType(value)
            else:
                r = RoleType(":".join(value.split("_", 1)))
        else:
            r = value
        if self.role and self.role != r:
            raise ValueError("The role property is immutable")
        if r:
            if r.name.startswith("endpoint") and self.resource:
                raise ValueError("Invalid role {} for collection".format(r.name))
            return r

    @hybrid_property
    def resource(self):
        return self._resource

    @resource.setter  # type: ignore
    def resource(self, value):
        self._resource = value

    @resource.expression  # type: ignore
    def resource(cls):
        return cls._resource

    @checks_property(prop="collection")
    def check_resource(self, value: uuid.UUID):
        if self.resource and self.resource != value:
            raise ValueError("The resource property is immutable")
        if value and self.role and self.role.name.startswith("endpoint"):
            raise ValueError("Invalid resource for endpoint role")
        if value:
            try:
                uuid.UUID(str(value))
            except Exception as e:
                current_app.logger.error(e)
                raise ValueError("Invalid resource format")

    # The following hybrid properties match names of properties in the
    # role#1.0.0 document which is used in the API.
    @hybrid_property
    def principal(self):
        """
        Property that describes the principal that the role is associated with
        in the form of a URN
        """
        principal_type = self.principal_type
        identity_id = self.identity_id
        if not (principal_type and identity_id):
            raise ValueError("Invalid value for principal property")

        if self.principal_type == PrincipalType.identity:
            return "urn:globus:auth:identity:" + identity_id
        elif self.principal_type == PrincipalType.group:
            return "urn:globus:groups:id:" + identity_id

    @principal.expression  # type: ignore
    def principal(self):
        return db.case(
            [
                (
                    Role.principal_type == "identity",
                    "urn:globus:auth:identity:" + Role.identity_id,
                ),
                (
                    Role.principal_type == "group",
                    "urn:globus:groups:id:" + Role.identity_id,
                ),
            ]
        )

    @principal.setter  # type: ignore
    def principal(self, value):
        if value:
            principal_type_prefix, identity_id = value.rsplit(":", 1)

            if principal_type_prefix == "urn:globus:auth:identity":
                principal_type = PrincipalType.identity
            elif principal_type_prefix == "urn:globus:groups:id":
                principal_type = PrincipalType.group
            else:
                principal_type = ""

            self.principal_type = principal_type
            self.identity_id = identity_id

    @checks_property(prop="principal")
    def check_principal(self, value: Optional[str]) -> None:
        old_principal_type = self.principal_type
        old_identity_id = self.identity_id

        if value:
            principal_type_prefix, identity_id = value.rsplit(":", 1)

            if principal_type_prefix not in [
                "urn:globus:auth:identity",
                "urn:globus:groups:id",
            ]:
                raise ValueError("Unknown principal type")

            if old_principal_type and (principal_type_prefix != old_principal_type):
                raise ValueError("Immutable field")
            if old_identity_id and identity_id != old_identity_id:
                raise ValueError("Immutable field")

            if identity_id != "UNAUTHENTICATED":
                try:
                    uuid.UUID(identity_id)
                except Exception as e:
                    current_app.logger.error(e)
                    raise ValueError("Invalid identity id")

    @hybrid_property
    def role_name(self):
        """
        Property that returns the unscoped name of the role that this object
        assigns to a user.
        """
        role = self.role
        if role:
            return role.name.split("_", 1)[1]

    @role_name.expression  # type: ignore
    def role_name(self):
        return db.case(
            [
                (Role.role.like("endpoint%"), func.substr(Role.role, 10)),
                (Role.role.like("collection%"), func.substr(Role.role, 12)),
            ]
        )

    @hybrid_property
    def is_collection_role(self):
        """
        Property that returns True if the role is for a collection
        """
        role = self.role
        if role:
            return role.name.split("_", 1)[0] == "collection"

    @is_collection_role.expression  # type: ignore
    def is_collection_role(cls):
        return cls.role.like("collection%")

    @hybrid_property
    def is_endpoint_role(self):
        """
        Property that returns True if the role is for an endpoint
        """
        role = self.role
        if role:
            return role.name.split("_", 1)[0] == "endpoint"

    @is_endpoint_role.expression  # type: ignore
    def is_endpoint_role(cls):
        return cls.role.like("endpoint%")

    def delete(self, collection_deleted=False):
        if self.primary_role:
            raise ValueError("Unable to delete a primary role")
        elif self.resource:
            transfer_endpoint_id = self.resource
        else:
            transfer_endpoint_id = Endpoint.query.one().endpoint_id

        if not collection_deleted:
            try:
                current_app.transfer_client.delete_endpoint_role(
                    transfer_endpoint_id, self.id
                )
            except globus_sdk.TransferAPIError as e:
                log_transfer_exception(e)
                if e.http_status != 404:
                    raise errors.BadGatewayError("Error deleting role")

        super().delete()

    def rollback(self, data=None):
        """
        Rolls back state to a previous version of the configuration_hash
        """
        if data:
            self.update(data["items"][0])

    def _create_role(self, transfer_endpoint_id: str) -> Optional[str]:
        try:
            if self.role_name == "owner":
                # For owner role, we let configuration_object create the role
                # id, since we don't express that role in transfer.
                return None
            else:
                res = current_app.transfer_client.add_endpoint_role(
                    transfer_endpoint_id, role_schema.dump(self)
                )
                return res["id"]
        except globus_sdk.TransferAPIError as e:
            log_transfer_exception(e)
            message = e.message  # noqa: B306
            if e.http_status == 409 and e.code == "Exists":
                raise errors.ResourceExistsError(error_description=message)
            elif e.http_status == 409:
                raise errors.ConflictError(error_description=message)
            elif e.http_status == 404:
                raise errors.NotFoundError(error_description=message)
            else:
                raise errors.BadGatewayError(error_description=message)

    def migrate_collection_uuid(self, collection: "Collection") -> None:
        """
        Used in v4 -> v5 migration, this method changes the resource associated
        with a role object to a new collection
        """
        self.resource = collection.id  # type: ignore
        self.refresh_configuration_hash()

    def __repr__(self):
        return "<Role {}>".format(self.id)
