from collections.abc import Mapping, Sequence
from typing import List, Optional, Union
import uuid

from sqlalchemy.ext.hybrid import hybrid_property

from . import db
from .configuration_object import ConfigurationObject
from .decorators import DocumentChecker, checks_property
from globus.manager.utils.path_restriction import PathRestrictions

from globus.manager.constants import ConfigurationObjectTag
from globus.manager.api.schemas.api import (
    path_restrictions_schema,
    sharing_policy_schema,
    PathRestrictionsDict,
    SharingPolicyLoadDict,
)


class SharingPolicy(ConfigurationObject, DocumentChecker):
    """
    Sharing policy model for mapped collection sharing restrictions. Currently,
    the sharing policy contains a set of sharing path restrictions associated
    with a list of users.
    """

    SHARING_POLICY_NAMESPACE = uuid.UUID("9cab22e8-2165-4d59-a9f0-6a0b4ce7af56")

    __tablename__ = "sharing_policy"
    __mapper_args__ = {
        "polymorphic_identity": ConfigurationObjectTag.SHARING_POLICY.name
    }
    configuration_schema = sharing_policy_schema
    _DATA_TYPE = "sharing_policy#1.0.0"
    _DATA_TYPES = [_DATA_TYPE]
    PROPERTIES = {_DATA_TYPE: ["collection_id", "sharing_restrict_paths", "users"]}

    id = db.Column(
        "id", db.TEXT, db.ForeignKey("configuration_object.id"), primary_key=True,
    )
    _collection_id = db.Column("collection_id", db.TEXT, db.ForeignKey("collection.id"))
    collection = db.relationship(
        "Collection", foreign_keys=[_collection_id], back_populates="sharing_policies"
    )

    __default_policy_id = None  # type: Optional[str]

    def __init__(
        self,
        doc: SharingPolicyLoadDict = None,
        collection=None,
        default_policy: bool = False,
    ):
        if doc:
            if "collection_id" not in doc:
                raise ValueError("Missing required collection_id property")
            if collection:
                self.collection = collection

            id = None  # type: Optional[str]
            if default_policy:
                id = self.default_collection_policy(str(doc["collection_id"]))
            elif "id" in doc:
                id = str(doc["id"])

            super().__init__(tag=ConfigurationObjectTag.SHARING_POLICY, id=id)
            self.post(doc)
        else:
            raise ValueError("Missing required parameter doc")

    @hybrid_property
    def collection_id(self) -> str:
        """
        The ID of the collection that this policy is associated with.
        """
        return self._collection_id

    @collection_id.expression  # type: ignore
    def collection_id(cls):
        return cls._collection_id

    @collection_id.setter  # type: ignore
    def collection_id(self, value: uuid.UUID):
        self._collection_id = str(value)
        self.configuration_hash["collection_id"] = str(value)

    @checks_property(prop="collection_id", required=True)
    def check_collection_id(self, value: uuid.UUID):
        try:
            uval = uuid.UUID(str(value))
            if uval and self.collection and str(uval) != self.collection.id:
                raise ValueError("Invalid collection_id property")
        except ValueError:
            raise ValueError("Invalid collection_id")

    @property
    def users(self) -> Optional[List[str]]:
        """
        The list of local usernames that this policy is associated with. This
        may be None, in which case this policy applies to all users.
        """
        value = self.configuration_hash.get("users")

        if value:
            return value
        else:
            return None

    @users.setter
    def users(self, value: Optional[List[str]]) -> None:
        if value:
            self.configuration_hash["users"] = [*value]
        else:
            self.configuration_hash["users"] = None

    @checks_property(prop="users")
    def check_users(self, value: Optional[List[str]]) -> None:
        if value is not None:
            if not isinstance(value, Sequence) or isinstance(value, str):
                raise ValueError("Invalid value for the users property.")
            for v in value:
                if not isinstance(v, str):
                    raise ValueError("Invalid value for the users property.")
        if self.is_default_policy and value:
            raise ValueError("Invalid value for the users property.")

    @property
    def sharing_restrict_paths(self) -> PathRestrictions:
        value = self.configuration_hash["sharing_restrict_paths"]
        return PathRestrictions(**value)

    @sharing_restrict_paths.setter
    def sharing_restrict_paths(self, value: Union[PathRestrictions, Mapping]):
        self.configuration_hash[
            "sharing_restrict_paths"
        ] = path_restrictions_schema.dump(value)

    @checks_property(prop="sharing_restrict_paths", required=True)
    def check_sharing_restrict_paths(self, value: PathRestrictionsDict):
        if isinstance(value, Mapping):
            sharing_restrict_paths = PathRestrictions(**value)
        else:
            raise ValueError("Invalid path restrictions document")

        all_path_restrictions = (
            sharing_restrict_paths.read
            + sharing_restrict_paths.read_write
            + sharing_restrict_paths.none
        )

        for pr in all_path_restrictions:
            for bad in ["$HOME", "~"]:
                if bad in pr:
                    raise ValueError(
                        "{} not allowed in sharing_restrict_paths".format(bad)
                    )

    def post(self, doc: SharingPolicyLoadDict) -> None:
        properties = []  # type: List[str]
        DATA_TYPE = doc.get("DATA_TYPE")
        if DATA_TYPE:
            properties = self.PROPERTIES.get(DATA_TYPE, [])

        if not properties:
            if DATA_TYPE and DATA_TYPE.startswith("sharing_policy#1."):
                properties = self.PROPERTIES[self._DATA_TYPE]
            else:
                raise ValueError("Unsupported DATA_TYPE {}".format(DATA_TYPE))

        for prop in properties:
            value = doc.get(prop)
            setattr(self, prop, value)

        self.check_doc(**doc)

        self.configuration_hash["DATA_TYPE"] = self._DATA_TYPE

    @property
    def is_default_policy(self) -> bool:
        if not self.__default_policy_id:
            self.__default_policy_id = self.default_collection_policy(
                self._collection_id
            )
        return self.__default_policy_id == self.id

    @staticmethod
    def default_collection_policy(collection_id: str) -> str:
        """
        Compute the well-known collection policy uuid for a given collection. This
        is used to migrate an existing collection's sharing_restrict_paths to the
        new sharing_policy format.

        :param collection_id:
            UUID string containing the collection's ID.

        :return str:
            Returns the uuid string of the collection policy
        """
        default_uuid = uuid.uuid5(SharingPolicy.SHARING_POLICY_NAMESPACE, collection_id)
        return str(default_uuid)

    def __repr__(self) -> str:
        return "<SharingPolicy {}>".format(self.id)
