from collections.abc import Sequence
import pkg_resources
import re
from typing import Dict, List, Mapping, Optional, Union
import uuid

from flask import current_app
from sqlalchemy import orm
from sqlalchemy.ext.hybrid import hybrid_property

from . import db, Account, ConfigurationObject
from .decorators import DocumentChecker, checks_property

import globus.manager.api.errors as errors
from globus.manager.constants import ConfigurationObjectTag
from globus.manager.api.schemas.api import (
    identity_mapping_schema,
    path_restrictions_schema,
    storage_gateway_private_schema,
)
from globus.manager.utils.path_restriction import PathRestrictions


class StorageGateway(ConfigurationObject, DocumentChecker):
    """
    Storage Gateway model

    This is the baseline definition of a storage gateways without any
    connector-specific policies (those are implemented in subclasses). This
    model is uses joined table inheritance with the configuration_object table
    to implement encryption of confidential data and sharing of the data
    across DTNs via AppSync.

    To dynamically resolve the connector ID to the proper concrete model
    implementation, the SQLAlchemy mapper uses the 'secondary_discriminator'
    from the base configuration_object class. This is set to the result of
    concatenating the strings 'STORAGE_GATEWAY' and the hex representation of
    the connector UUID.
    """

    __tablename__ = "storage_gateway"

    _DATA_TYPES = ["storage_gateway#1.0.0", "storage_gateway#1.1.0"]

    CONNECTOR_POLICIES = {}  # type: Mapping[str, List[str]]
    """
    Connector-specific mapping of DATA_TYPE to list of properties which can
    be present in the policy document
    """

    connector_policy_names = []  # type: List[str]

    configuration_schema = storage_gateway_private_schema

    storage_gateway_id = db.Column(
        "id", db.TEXT, db.ForeignKey("configuration_object.id"), primary_key=True
    )

    _display_name = db.Column("display_name", db.TEXT, nullable=False)

    _high_assurance = db.Column(
        "require_high_assurance", db.BOOLEAN, nullable=False, server_default="0"
    )

    connector_id = db.Column("connector_id", db.TEXT, nullable=False)

    _identity_mappings = None
    _restrict_paths = None  # type: Optional[PathRestrictions]

    def __init__(self, appsync_data=None, doc=None):

        if appsync_data is not None:
            args = {
                **appsync_data,
                "tag": ConfigurationObjectTag.STORAGE_GATEWAY,
                "secondary_discriminator": self.CONNECTOR_ID,
            }
            super().__init__(**args)
            self.init_appsync()
        elif doc is not None and (doc.get("DATA_TYPE") in self._DATA_TYPES):
            if doc["connector_id"] != self.CONNECTOR_ID:
                raise ValueError("Invalid connector")

            super().__init__(
                id=uuid.uuid4(),
                tag=ConfigurationObjectTag.STORAGE_GATEWAY,
                secondary_discriminator=self.CONNECTOR_ID,
            )

            self.storage_gateway_id = self.id
            self.check_doc(**doc)
            self._put(**doc)
            self.refresh_configuration_hash()
            self._check_required()
        else:
            raise ValueError("initer Unsupported DATA_TYPE")

    @staticmethod
    def replace_path_vars(path, username=None, home_directory=None):
        """
        Replace all instances of `$HOME` and `~` with `home_directory` and all
        instances of `$USER` with `username` in path.

        :param path:
            String containing the path to rewrite.
        :param username:
            The string to replace literal `$USER` in the paths
        :param home_directory:
            The string to replace literal `$HOME` in the paths
        :returns:
            Returns the value of `path`, with all instances of `$HOME and, `~`,
            and `$USER` replaced.
        """

        def replace_home_and_user(match):
            m = match.groups()[0]

            if m == "$HOME" or m == "~":
                return home_directory
            elif m == "$USER":
                return username

        return re.sub(r"(\$HOME|\$USER|~)", replace_home_and_user, path)

    def absolute_restrict_paths(self, username=None):
        """
        Call :py:code:`StorageGateway.replace_path_vars()` on each path in the
        path restrictions for this and return a new PathRestrictions object
        with the new paths.

        :param username:
            Connector-specific username to use for $USER.
        """
        home_directory = self.home_directory(username=username)
        if self._restrict_paths:
            absolute_paths = PathRestrictions()
            absolute_paths.read = [
                self.replace_path_vars(
                    r, username=username, home_directory=home_directory
                )
                for r in (self.restrict_paths.read or [])
            ]
            absolute_paths.read_write = [
                self.replace_path_vars(
                    r, username=username, home_directory=home_directory
                )
                for r in (self.restrict_paths.read_write or [])
            ]
            absolute_paths.none = [
                self.replace_path_vars(
                    r, username=username, home_directory=home_directory
                )
                for r in (self.restrict_paths.none or [])
            ]
            return absolute_paths

    def init_appsync(self):
        if self.configuration_hash:
            self._put(**self.configuration_hash)
            self.refresh_configuration_hash()

    @hybrid_property
    def display_name(self):
        return self._display_name

    @display_name.setter  # type: ignore
    def display_name(self, value):
        if value != self._display_name:
            self._display_name = value
            self.configuration_hash["display_name"] = value

    @checks_property(prop="display_name")
    def _check_display_name(self, value: str) -> None:
        if not isinstance(value, str):
            raise ValueError()

    @property
    def endpoint(self):
        return ConfigurationObject.query.filter(
            ConfigurationObject.tag == ConfigurationObjectTag.ENDPOINT
        ).one()

    @classmethod
    def polymorphic_class(cls, discriminator):
        return ConfigurationObject.polymorphic_class(
            ConfigurationObjectTag.STORAGE_GATEWAY, discriminator
        )

    @hybrid_property
    def high_assurance(self):
        return self._high_assurance

    @high_assurance.setter  # type: ignore
    def high_assurance(self, value):
        old_value = self._high_assurance
        if old_value is not None and old_value != value:
            raise ValueError("The high_assurance property is immutable")
        elif value is None or not isinstance(value, bool):
            raise ValueError("Invalid high_assurance value")

        self._high_assurance = bool(value)
        self.configuration_hash["high_assurance"] = bool(value)

    @high_assurance.deleter  # type: ignore
    def high_assurance(self):
        current_app.logger.debug("High assurance deleter called")
        raise AttributeError("The high_assurance property is immutable")

    @orm.validates("storage_gateway_id")
    def __immutable_property(self, key, value):
        old_value = getattr(self, key, None)
        if old_value is not None and old_value != value:
            raise ValueError("The {} property is immutable".format(key))
        return value

    @checks_property(prop="connector_id")
    def check_connector_id(self, value: str):
        old_value = self.connector_id
        if old_value is not None and old_value != value:
            raise AttributeError()

    allowed_domains = ConfigurationObject.model_property(
        "allowed_domains", required=True
    )

    @checks_property(prop="allowed_domains")
    def check_allowed_domains(self, value: List[str]):
        current_app.logger.debug("check_allowed_domains called")
        if not value or not isinstance(value, list):
            raise ValueError("Invalid allowed_domains value")
        for v in value:
            if not isinstance(v, str):
                raise ValueError("Invalid allowed_domains value")
        invalid_domains = set(value)
        idps = current_app.auth_client.get(
            "/v2/api/identity_providers", query_params={"domains": ",".join(value)}
        )
        for idp in idps["identity_providers"]:
            for domain in idp["domains"]:
                if idp["name"] == "Email Address":
                    # Ignore email domain, as it can't be used for login as it
                    # has no authentication time associated with it
                    continue
                if domain in invalid_domains:
                    invalid_domains.remove(domain)
                if not invalid_domains:
                    break

        if invalid_domains:
            raise ValueError(
                "Globus Auth does not have an identity provider registration for "
                "the following domains, so they can not be used with Globus "
                "Connect Server: {}".format(", ".join(invalid_domains))
            )

    require_mfa = ConfigurationObject.model_property("require_mfa", required=True)

    @require_mfa.getter  # type: ignore
    def require_mfa(self) -> bool:
        return self.configuration_hash.get("require_mfa", False)

    @property
    def identity_mappings(self):
        if self._identity_mappings is None:
            if "identity_mappings" in self.configuration_hash:
                identity_mappings = []
                value = self.configuration_hash["identity_mappings"]
                for i in value:
                    if isinstance(i, dict):
                        selected_identity_mapping = i["DATA_TYPE"].split("#", 1)[0]
                        ep = next(
                            pkg_resources.iter_entry_points(
                                "globus.manager.api.identity_mapping",
                                selected_identity_mapping,
                            )
                        )
                        entry_point = ep.load()
                        try:
                            mapper = entry_point.from_json(
                                i, connector=self.connector_id, storage_gateway=self.id
                            )
                        except ValueError as e:
                            raise ValueError(
                                "Invalid identity_mappings: {}".format(str(e))
                            )
                        identity_mappings.append(mapper)
                    else:
                        identity_mappings.append(i)
                self._identity_mappings = identity_mappings

        value = self._identity_mappings
        if value is None:
            raise AttributeError("identity_mappings")
        return value

    @identity_mappings.setter
    def identity_mappings(self, value):
        if value is None:
            if "identity_mappings" in self.configuration_hash:
                del self.configuration_hash["identity_mappings"]
        else:
            identity_mappings = []
            for i in value:
                json_value = identity_mapping_schema.dump(i)
                identity_mappings.append(json_value)
            self.configuration_hash["identity_mappings"] = identity_mappings
        self._identity_mappings = None

    @identity_mappings.deleter
    def identity_mappings(self):
        self.identity_mappings = None

    authentication_timeout_mins = ConfigurationObject.model_property(
        "authentication_timeout_mins"
    )

    @authentication_timeout_mins.setter  # type: ignore
    def authentication_timeout_mins(self, value: Optional[int]) -> None:
        if value is None or not isinstance(value, int) or value <= 0:
            # default to 11 days
            self.configuration_hash["authentication_timeout_mins"] = 11 * 24 * 60
        else:
            self.configuration_hash["authentication_timeout_mins"] = int(value)

    @authentication_timeout_mins.deleter  # type: ignore
    def authentication_timeout_mins(self, value: Optional[int]) -> None:
        if value is None or not isinstance(value, int) or value <= 0:
            # default to 11 days
            self.configuration_hash["authentication_timeout_mins"] = 11 * 24 * 60
        else:
            self.configuration_hash["authentication_timeout_mins"] = int(value)

    @checks_property(prop="authentication_timeout_mins")
    def check_authentication_timeout_mins(self, value: Optional[int]):
        if self.high_assurance:
            if value is None or not isinstance(value, int) or value <= 0:
                raise ValueError("Invalid authentication_timeout_mins value")
        elif isinstance(value, int):
            if value <= 0:
                raise ValueError("Invalid authentication_timeout_mins value")
        elif value is not None:
            raise ValueError("Invalid authentication_timeout_mins value")

    @property
    def restrict_paths(self) -> PathRestrictions:
        if self._restrict_paths is None:
            if "restrict_paths" in self.configuration_hash:
                if self.configuration_hash["restrict_paths"]:
                    self._restrict_paths = PathRestrictions(
                        **self.configuration_hash["restrict_paths"]
                    )
                else:
                    self._restrict_paths = None
        value = self._restrict_paths
        if value is None:
            raise AttributeError("restrict_paths")

        return value

    @restrict_paths.setter
    def restrict_paths(self, value: Union[Dict, PathRestrictions]):
        if value != self.configuration_hash.get("restrict_paths"):
            if value:
                self.configuration_hash[
                    "restrict_paths"
                ] = path_restrictions_schema.dump(value)
                self._restrict_paths = None
            else:
                self.configuration_hash.pop("restrict_paths")
                self._restrict_paths = None

    @restrict_paths.deleter
    def restrict_paths(self):
        self.restrict_paths = None

    load_dsi_module = ConfigurationObject.model_property("load_dsi_module")

    @checks_property(prop="load_dsi_module")
    def check_load_dsi_module(self, value: str):
        old_value = getattr(self, "load_dsi_module", None)
        if old_value and value and value != old_value:
            raise AttributeError()

    process_user = ConfigurationObject.model_property("process_user")

    @checks_property(prop="process_user")
    def check_process_user(self, value: str):
        old_value = getattr(self, "process_user", None)
        if old_value and value and value != old_value:
            raise AttributeError()

    @property
    def require_credential(self):
        return True

    @staticmethod
    def __validate_string_list(name: str, value: Optional[List[str]]):
        if value:
            if not isinstance(value, Sequence):
                raise ValueError()
            for v in value:
                if not isinstance(v, str):
                    raise ValueError()

    users_allow = ConfigurationObject.model_property("users_allow")

    @checks_property(prop="users_allow")
    def _check_users_allow(self, value: Optional[List[str]]):
        self.__validate_string_list("users_allow", value)

    users_deny = ConfigurationObject.model_property("users_deny")

    @checks_property(prop="users_deny")
    def _check_users_deny(self, value: Optional[List[str]]) -> None:
        self.__validate_string_list("users_deny", value)

    def default_mapping(self, identity_data):
        pass

    def map_identities(self, identity_data):
        """
        Apply the configured identity mappings to an authorized identity set.
        Mappings are applied in the order of the mapping list in the
        Storage Gateway configuration document.

        For each mapping, all identities are attempted to be matched in order
        by their identity ID. If there is mapping, the mapped username is
        checked against the Storage Gateway allowed and denied user policy, and
        if allowed added the the result list.

        :param identity_data:
            An IdentitySet containing the list of Auth identity resources which
            are authorized by the storage gateway policies.

        :returns:
            This function returns a list of Account objects containing the
            mapping information for all the identities in identity_data.
        """
        current_app.logger.debug("Mapping %s", identity_data)
        self._check_required()

        unmapped_identities = sorted([i for i in identity_data], key=lambda x: x["id"])
        mappings = []
        # If no mappings are defined, use self, which has a map_identity which
        # will fall back to self.default_mapping
        identity_mappings = getattr(self, "identity_mappings", [self])

        mapped_but_invalid = set()

        for m in identity_mappings:
            for i in unmapped_identities:
                # Do not map 'closed' accounts.

                # Note: globusid.org once allowed 'private' identities. That feature is now
                # disabled, only legacy accounts can have 'private' status. globusid accounts
                # can not be 'unused' or 'closed'. So we need not worry about 'private' being
                # 'closed'.
                if i["status"] == "closed":
                    continue
                try:
                    mapping = m.map_identity(i)
                    if mapping:
                        if self.check_user_allowed(mapping):
                            mappings.append(
                                Account(
                                    identity_id=i["id"],
                                    username=mapping,
                                    storage_gateway_id=self.id,
                                )
                            )
                        else:
                            mapped_but_invalid.add(mapping)
                except Exception as e:
                    current_app.logger.warning(
                        "Mapping %s failed for %s: %s", m, identity_data, str(e)
                    )
        current_app.logger.debug("Mapped to %s", mappings)

        if not mappings and mapped_but_invalid:
            raise errors.InvalidUserError(usernames=sorted(list(mapped_but_invalid)))

        return mappings

    def map_identity(self, identity_data):
        current_app.logger.debug("Mapping %s", identity_data)
        try:
            for m in self.identity_mappings:
                try:
                    mapping = m.map_identity(identity_data)
                    if mapping:
                        return mapping
                except Exception as e:
                    current_app.logger.debug(
                        "Mapping failed for %s: %s", identity_data, e
                    )
            return None
        except AttributeError:
            self._check_required()
            return self.default_mapping(identity_data)

    def check_user_allowed(self, username):
        """
        Check whether a user is allowed to access a collection, based on a
        storage_gateway's policies.
        """
        try:
            users_allow = self.users_allow
        except AttributeError:
            users_allow = None
        try:
            users_deny = self.users_deny
        except AttributeError:
            users_deny = None

        if users_deny and username in users_deny:
            # Explicitly denied
            return False
        if users_allow:
            if username in users_allow:
                # Explicitly allowed
                return True
            else:
                # Others are explicitly allowed, this one is not
                return False
        else:
            # Not denied and nobody explicitly allowed: implicitly allowed
            return True

    def _check_required(self):
        if not getattr(self, "allowed_domains", None):
            raise ValueError("Missing required allowed_domains")

        if len(self.allowed_domains) > 1 and not getattr(
            self, "identity_mappings", None
        ):
            raise ValueError(
                "Storage Gateway must include identity_mappings "
                "if more than one domain is allowed"
            )

        if self.require_mfa and not self.high_assurance:
            raise ValueError("Only HA gateways support the MFA requirement.")

    def put(self, **doc):
        self.check_doc(**doc)
        self._put(**doc)
        self._check_required()

    def check_doc(self, **doc):
        super().check_doc(**doc)
        if doc.get("high_assurance", False) and not doc.get(
            "authentication_timeout_mins"
        ):
            raise ValueError(
                "Missing required authentication_timeout_mins for a high assurance storage gateway"
            )

        policies = {**doc.get("policies", {})}

        if policies:
            DATA_TYPE = policies.pop("DATA_TYPE", None)
            if DATA_TYPE not in self.POLICY_DATA_TYPES:
                raise ValueError("Unsupported DATA_TYPE {}".format(DATA_TYPE))
            for policy in policies:
                if policy not in self.connector_policy_names:
                    raise ValueError("Unsupported policy {}".format(policy))

    def _put(self, DATA_TYPE=None, **doc):
        if DATA_TYPE not in self._DATA_TYPES:
            raise ValueError("Unsupported DATA_TYPE")
        policies = doc.pop("policies", {})

        self.connector_id = doc["connector_id"]
        self.high_assurance = doc.get("high_assurance", False)
        self.authentication_timeout_mins = doc.get("authentication_timeout_mins")
        self.require_mfa = doc.get("require_mfa", False)
        self.display_name = doc.get("display_name")
        self.allowed_domains = doc.get("allowed_domains")
        self.restrict_paths = doc.get("restrict_paths")
        self.identity_mappings = doc.get("identity_mappings")
        self.users_allow = doc.get("users_allow")
        self.users_deny = doc.get("users_deny")
        self.deleted = doc.get("deleted", False)

        if policies:
            self._put_policies(**policies)

    def patch(self, DATA_TYPE=None, policies=None, **doc):
        """
        Apply PATCH semantics to this Storage Gateway.
        This method assumes the patch has been validated against the input
        schema, and that any constraints are handled by the individual property
        setters or sqlalchemy validates hooks. This method applies changes to
        the policies subdocument by calling the patch_policies method.
        """
        if DATA_TYPE not in self._DATA_TYPES:
            raise ValueError("Unsupported DATA_TYPE")

        props = [
            "connector_id",
            "high_assurance",
            "require_mfa",
            "authentication_timeout_mins",
            "display_name",
            "allowed_domains",
            "restrict_paths",
            "identity_mappings",
            "users_allow",
            "users_deny",
        ]

        full_patch = {**doc, **{"policies": policies}}
        self.check_patch_doc(**full_patch)

        for prop in props:
            if prop in doc:
                if doc[prop] is None:
                    delattr(self, prop)
                else:
                    setattr(self, prop, doc[prop])
        self._check_required()

        if policies:
            self.patch_policies(**policies)
        self.refresh_configuration_hash()

    @property
    def policies(self):
        return self.configuration_schema.dump(self)["policies"]

    @policies.setter
    def policies(self, value):
        self.put_policies(**value)

    def home_directory(self, username):
        """
        Look up the connector-specific home directory of the requested user.

        :param username:
            Name of the user look up the home directory of.
        :returns:
            If the connector supports the concept of a home directory, this
            returns the home directory of the user. If the connector does not,
            this function returns '/'
        """
        return "/"

    def put_policies(self, DATA_TYPE=None, **policies):
        """
        Replace the entire policy document with the values in policies.

        If a policy not in the Connector subclass's connector_policy_names is
        passed in, raise ValueError.

        If a policy in the Connector subclass's connector_policy_names is not
        passed in, set to None.
        """
        if DATA_TYPE not in self.POLICY_DATA_TYPES:
            raise ValueError("Unsupported DATA_TYPE")

        for p in policies:
            if p not in self.connector_policy_names:
                raise ValueError("Unsupported policy {}".format(p))
        self._put_policies(DATA_TYPE=DATA_TYPE, **policies)

    def _put_policies(self, DATA_TYPE=None, **policies):
        for p in self.connector_policy_names:
            if p not in policies:
                policies[p] = None

        for p in policies:
            if p in self.connector_policy_names:
                setattr(self, p, policies[p])

    def patch_policies(self, **policies):
        """
        Apply PATCH semantics to the policies of this Storage Gateway.
        Order of application of policies is undefined; subclasses can override
        this if there are some dependencies.
        """
        DATA_TYPE = policies.pop("DATA_TYPE", None)
        current_app.logger.error("patch_policies: %s", policies)

        if DATA_TYPE not in self.POLICY_DATA_TYPES:
            raise ValueError("Unsupported DATA_TYPE")

        for p in policies:
            if p not in self.connector_policy_names:
                raise ValueError("Unsupported policy {}".format(p))

        for p, v in policies.items():
            current_app.logger.error("Setting %s on %s to %s", p, self, v)
            if v is None:
                delattr(self, p)
            else:
                setattr(self, p, v)

    def rollback(self, data=None):
        if data:
            self.update(data["items"][0])
        else:
            self.configuration_hash = (
                db.session.query(ConfigurationObject.configuration_hash)
                .filter(ConfigurationObject.id == self.id)
                .one()[0]
            )
            self.init_appsync()

    def __repr__(self):
        return "<StorageGateway {}>".format(self.display_name)
