import json
import socket
import urllib.parse as urlparse

from flask import current_app
import paho.mqtt.client

# Override TCP-based compatibility hack with the real socketpair
# function
paho.mqtt.client._socketpair_compat = socket.socketpair


class MqttConnectionInfo:
    """
    This class holds the the value from one of the  mqttConnections properties
    returned as part of the AppSync subscription extension
    """

    def __init__(self, url, topics, client):
        # URL of the websocket broker
        self.url = url
        # list of topics available at this url
        self.topics = topics
        # client to pass to mqtt broker. This allows us to reconnect
        # and resume notifications where they left off
        self.client = client


class SubscriptionFieldInfo:
    """
    This class holds the the value from one of the newSubscriptions properties
    returned as part of the AppSync subscription extension
    """

    def __init__(self, topic=None, expireTime=None):
        # Topic is the opaque topic string
        self.topic = topic
        # expireTime is when this topic expires (not implemented AFAICT)
        self.expire_time = expireTime


class SubscriptionInfo:
    """
    This class holds the the parsed result from parsing the subscription
    extension returned from AppSync.
    """

    def __init__(self, extension):
        sub = extension
        self.mqtt_connections = [
            MqttConnectionInfo(**s) for s in sub["mqttConnections"]
        ]
        self.new_subscriptions = {
            field: SubscriptionFieldInfo(**data)
            for field, data in sub["newSubscriptions"].items()
        }


class SubscriptionConnection:
    """
    This class implements a single connection to the mqtt broker,
    with callbacks for on_message and on_disconnect
    """

    def __init__(self, url, client, topics, on_message, on_disconnect):
        mqttUrl = urlparse.urlsplit(url)
        mqttc = paho.mqtt.client.Client(
            client_id=client,
            clean_session=False,
            transport="websockets",
            userdata=current_app._get_current_object(),
        )
        self.topics = topics
        mqttc.tls_set()
        mqttc.on_message = self.on_message
        mqttc.on_connect = self.on_connect
        mqttc.on_disconnect = self.on_disconnect
        mqttc.enable_logger(current_app.logger)
        mqttc.ws_set_options(
            path="{}?{}".format(mqttUrl.path, mqttUrl.query),
            headers={"Host": mqttUrl.netloc},
        )

        current_app.logger.debug("Connecting to %s for real-time data", mqttUrl.netloc)
        mqttc.connect(mqttUrl.netloc, 443, 60)
        mqttc.loop_start()

        self.mqttc = mqttc
        self._on_message = on_message
        self._on_disconnect = on_disconnect

    def on_connect(self, client, userdata, flags, rc, properties=None):
        with userdata.app_context():
            current_app.logger.debug("Connected to appsync")
            self.mqttc.subscribe(list([(t, 1) for t in self.topics]))

    def disconnect(self):
        self.mqttc.disconnect()
        self.mqttc.loop_stop()

    def on_message(self, mqttc, userdata, msg):
        context = userdata.app_context()
        with context:
            current_app.logger.debug("Message reeived from appsync")
            payload = msg.payload.decode("utf8")
            if self._on_message:
                try:
                    self._on_message(context, json.loads(payload))
                except Exception as e:
                    current_app.logger.debug("Exception processing message %s", str(e))

    def on_disconnect(self, mqttc, userdata, rc):
        with userdata.app_context():
            current_app.logger.debug("Disconnect received from appsync")
            if self._on_disconnect:
                self._on_disconnect(rc)


class Subscription:
    def __init__(self, info, on_message=None, on_disconnect=None):
        self.info = info
        self.connections = {}

        topics = set()
        topics_for_url = {}
        client_for_url = {}

        # Determine what topics we are trying to subscribe to
        for subf in info.new_subscriptions.values():
            topic = subf.topic
            topics.add(topic)

        # Determine which URLs provide those topics
        for c in info.mqtt_connections:
            url = c.url
            client = c.client
            url_topics = set()
            for t in c.topics:
                if t in topics:
                    url_topics.add(t)
            if len(url_topics) > 0:
                topics_for_url[url] = url_topics
                client_for_url[url] = client

        # Connect to those URLs
        for url in topics_for_url:
            client = client_for_url[url]
            self.connections[client] = SubscriptionConnection(
                url,
                client_for_url[url],
                list(topics_for_url[url]),
                on_message,
                on_disconnect,
            )

    def disconnect(self):
        for c in self.connections.values():
            c.disconnect()
