from flask import Flask
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import event
from sqlalchemy.engine import Engine
from sqlite3 import Connection as SQLite3Connection
import sys
import os
import csv
import json

db = SQLAlchemy()


class Session(db.Model):  # type: ignore
    __tablename__ = "session"
    id = db.Column(db.Text, primary_key=True)
    ts = db.Column(db.DateTime)
    client_addr = db.Column(db.Text)
    auth_type = db.Column(db.Text)
    process_user = db.Column(db.Text)
    storage_user = db.Column(db.Text)
    root_path = db.Column(db.Text)
    task_id = db.Column(db.Text)
    collection_id = db.Column(db.Text)
    identity_id = db.Column(db.Text)
    operations = db.relationship("Operation", back_populates="session")


class Operation(db.Model):  # type: ignore
    __tablename__ = "operation"
    session_id = db.Column(db.Text, db.ForeignKey("session.id"), primary_key=True)
    op = db.Column(db.Text, primary_key=True)
    start_ts = db.Column(db.DateTime, primary_key=True)
    end_ts = db.Column(db.DateTime, primary_key=True, nullable=True)
    path = db.Column(db.Text, primary_key=True)
    peer_addr = db.Column(db.Text)
    length = db.Column(db.BigInteger)
    result = db.Column(db.Text)
    count = db.Column(db.Integer)
    session = db.relationship("Session", back_populates="operations")


def create_app(db_path):
    app = Flask(__name__, instance_relative_config=False)
    app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite:///{}".format(db_path)
    app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False

    db.init_app(app)
    with app.app_context():
        db.create_all()
        return app


# If the underlying database is sqlite, turn on
# foreign key support for the connection.
@event.listens_for(Engine, "connect")
def _set_sqlite_pragma(dbapi_connection, connection_record):
    if isinstance(dbapi_connection, SQLite3Connection):
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA foreign_keys=ON")
        cursor.close()


AUDIT_FIELDS = [
    "start_ts",
    "end_ts",
    "op",
    "path",
    "peer_addr",
    "length",
    "result",
    "count",
    "client_addr",
    "auth_type",
    "process_user",
    "storage_user",
    "root_path",
    "task_id",
    "collection_id",
    "identity_id",
]
AUDIT_OPS = ["STOR", "RETR", "STAT", "CKSM", "DELE", "REN", "MKD", "RMD", "RDEL"]


def format_result(result, format=None, fields=None):
    if format == "csv":
        fieldnames = fields
        if not fields:
            fieldnames = AUDIT_FIELDS
        csvwriter = csv.DictWriter(sys.stdout, fieldnames, lineterminator=os.linesep)
        csvwriter.writeheader()
    elif format == "json":
        print("[", end="")

    comma = ""
    for r in result:
        data = {
            "op": r.op,
            "path": r.path,
            "start_ts": str(r.start_ts),
            "end_ts": str(r.end_ts),
            "peer_addr": r.peer_addr,
            "length": r.length,
            "result": r.result,
            "count": r.count,
            "client_addr": r.session.client_addr,
            "auth_type": r.session.auth_type,
            "process_user": r.session.process_user,
            "storage_user": r.session.storage_user,
            "root_path": r.session.root_path,
            "task_id": r.session.task_id,
            "collection_id": r.session.collection_id,
            "identity_id": r.session.identity_id,
        }

        if format == "json":
            print(comma)
            comma = ","
            if not fields:
                print(json.dumps(data, sort_keys=True), end="")
            else:
                print(
                    json.dumps(
                        dict((key, data[key]) for key in fields if key in data),
                        sort_keys=True,
                    ),
                    end="",
                )
        elif format == "csv":
            csvwriter.writerow(data)
        else:
            if not fields:
                fields = [
                    "op",
                    "path",
                    "start_ts",
                    "collection_id",
                    "identity_id",
                    "storage_user",
                    "auth_type",
                    "result",
                ]
            print("   ".join("{}={}".format(key, data[key]) for key in fields))

    if format == "json":
        print("]")
