import json
import os

import click
from flask import current_app

from globus.manager.api.app import make_app
import globus.manager.api.models as m
from globus.manager.scripts.utils.setup_utils import load_keychain
from globus.manager.utils.keychain import NodeKey
from gcs_cli.parsing.command_decorator import globus_command
from globus.manager.constants import ConfigurationObjectTag
from globus.manager.api.blueprints.utils import patch_object


@globus_command(
    name="upgrade",
    short_help="Apply policy changes after package update",
    disable_options=["format", "explicit_host"],
)
@click.option("--verbose", "-v", is_flag=True)
@click.pass_context
def endpoint_upgrade(ctx, verbose,) -> None:
    """
    Apply policy changes after package updates.
    """
    if os.getuid() != 0:
        click.echo("This command must be run as the root user", err=True)
        ctx.exit(1)

    if verbose:
        os.environ["GCS_MANAGER_LOG_LEVEL"] = "DEBUG"

    app = make_app(app_config="globus.manager.api.config.endpoint_upgrade")
    m.db.init_app(app)

    with app.app_context():
        app.logger.info("endpoint_upgrade: start")
        m.db.create_all()
        m.db.session.commit()

        upgrade = EndpointUpgrade()
        upgrade.run()
        app.logger.info("endpoint_upgrade: end")


def lookup_config(column):
    field = m.Configuration.query.filter_by(name=column).one_or_none()
    if field:
        field = field.value
    return field


class EndpointUpgrade:
    def __init__(self):
        environment = lookup_config("environment")
        if environment == "default":
            environment = "production"
        current_app.config["ENVIRONMENT"] = environment
        current_app.config["CLIENT_ID"] = lookup_config("client_id")
        current_app.config["CLIENT_SECRET"] = lookup_config("client_secret")
        current_app.config["NODE_KEY"] = lookup_config("node_key")
        if current_app.config["NODE_KEY"]:
            current_app.config["NODE_KEY"] = NodeKey(
                **json.loads(current_app.config["NODE_KEY"])
            )

        if (
            current_app.config["CLIENT_ID"] is None
            or current_app.config["CLIENT_SECRET"] is None
            or current_app.config["NODE_KEY"] is None
        ):
            click.echo("This node is not configured... exiting", err=True)
            click.get_current_context().exit(1)

        load_keychain()

    def run(self):
        tasks = [
            self.update_guest_collections,
            self.update_deleted_collections,
        ]

        for task in tasks:
            task()

    def update_guest_collections(self):
        """
        [5.4.17] If mapped.force_encryption=True, propagate to guest collection.
        [5.4.17] Update guest Transfer docs with mapped inheritance.
        """
        items = current_app.appsync_client.query_all(
            tag=ConfigurationObjectTag.COLLECTION
        )
        for item in items:
            item["configuration_hash"] = current_app.keychain.decrypt(
                item["configuration"]
            )

        encrypted_mapped_ids = [
            i["id"]
            for i in items
            if i["configuration_hash"].get("collection_type") == "mapped"
            and i["configuration_hash"].get("force_encryption") is True
            and not i["configuration_hash"].get("deleted") is True
        ]

        guests_to_update = [
            i
            for i in items
            if i["configuration_hash"].get("collection_type") == "guest"
            and not i["configuration_hash"].get("deleted") is True
        ]

        def _update_guest_collection(guest):
            model = current_app.appsync_client.process_item(guest)
            if (
                guest["configuration_hash"].get("mapped_collection_id")
                in encrypted_mapped_ids
                and guest["configuration_hash"].get("force_encryption") is not True
            ):
                patch_document = {
                    "DATA_TYPE": guest["configuration_hash"].get("DATA_TYPE"),
                    "force_encryption": True,
                }
                patch_object(model, patch_document)
            else:
                model.update_transfer_endpoint()

        for guest in guests_to_update:
            click.echo("Updating guest collection {}.".format(guest["id"]))
            try:
                _update_guest_collection(guest)
            except Exception as e:
                click.echo(
                    "Could not update guest collection {}, {}".format(
                        guest["id"], str(e)
                    ),
                    err=True,
                )

    def update_deleted_collections(self):
        """
        [5.4.36] Fixup domain_name and domain on deleted collections
        """
        # Get all collection items from appsync
        items = current_app.appsync_client.query_all(
            tag=ConfigurationObjectTag.COLLECTION
        )

        # Decrypt each item's configuration hash
        def _decrypt(i):
            i["configuration_hash"] = current_app.keychain.decrypt(i["configuration"])
            return i

        items = map(_decrypt, items)

        # Filter the collection list
        def _filter(i):
            # Only operate on deleted collections
            if not i["configuration_hash"].get("deleted"):
                return False

            if i["configuration_hash"].get("domain"):
                return True

            expected_domain_name = i["id"] + "-[DELETED]"
            if i["configuration_hash"].get("domain_name") != expected_domain_name:
                return True

            return False

        items = filter(_filter, items)

        # Finally, process the broken collections
        for item in items:
            click.echo("Updating deleted collection {}.".format(item["id"]))
            model = current_app.appsync_client.process_item(item)
            model.domain = None
            model.domain_name = item["id"] + "-[DELETED]"
            current_app.appsync_client.update_object(model)
            m.db.session.add(model)
            m.db.session.commit()
