import os
import json

import click
from flask import current_app

import globus.manager.api.models as models
from globus.manager.api.app import make_app
from gcs_cli.parsing.command_decorator import globus_command
from globus.manager.scripts.utils.setup_utils import load_keychain
from globus.manager.utils.keychain import NodeKey


@globus_command(
    name="delete",
    short_help="Delete an OIDC server from this endpoint",
    disable_options=["format", "explicit_host"],
    hidden=True,
    help="""\
The OIDC delete command deletes an OIDC server configuration or registration
from the endpoint.  The domain used by a managed OIDC server configuration can
not be reused.

== EXAMPLE

globus-connect-server oidc delete

""",
)
def oidc_delete() -> None:
    if os.getuid() != 0:
        click.echo("This command must be run as the root user", err=True)
        click.get_current_context().exit(1)

    app = make_app(app_config="globus.manager.api.config.oidc_cli")
    models.db.init_app(app)

    with app.app_context():
        models.db.create_all()
        models.db.session.commit()

        # Needed by AppSync
        for conf in models.Configuration.query:
            if conf.name == "node_key":
                value = NodeKey(**json.loads(conf.value))
            else:
                value = conf.value
            app.config[conf.name.upper()] = value

        # Make sure `node setup` has been run
        if "CLIENT_ID" not in app.config:
            raise Exception("This node has not been setup.")

        try:
            # Load the keychain
            load_keychain()

            # Get the endpoint
            endpoint_model = models.Endpoint.query.one()

            if endpoint_model.has_oidc_server:
                endpoint_model.oidc_server.delete()
                endpoint_model.oidc_server.write_config()
                endpoint_model.oidc_server = None
                endpoint_model.refresh_configuration_hash()
                current_app.appsync_client.update_object(endpoint_model)
                models.db.session.add(endpoint_model)
                models.db.session.commit()
                click.echo("OIDC Server removed from this endpoint.")

            elif endpoint_model.has_oidc_registration:
                endpoint_model.oidc_registration.delete()
                endpoint_model.oidc_registration = None
                endpoint_model.refresh_configuration_hash()
                current_app.appsync_client.update_object(endpoint_model)
                models.db.session.add(endpoint_model)
                models.db.session.commit()
                click.echo("OIDC Registration removed from this endpoint.")

            else:
                click.echo("No OIDC configuration found.")
                click.get_current_context().exit(1)

        except Exception as e:
            click.echo("Error while deleting OIDC server: {}".format(str(e)))
