import os
import json
import time

import click
from flask import current_app
import requests

import globus.manager.api.models as models
from globus.manager.api.app import make_app
from gcs_cli.parsing.command_decorator import globus_command
from globus.manager.scripts.utils.setup_utils import load_keychain
from globus.manager.scripts.utils.dns_patterns import verify_domain_name_option
from globus.manager.utils.keychain import NodeKey


@globus_command(
    name="register",
    short_help="Register an existing OIDC server for use by this endpoint",
    disable_options=["format", "explicit_host"],
    help="""\
The OIDC register command registers an existing OIDC server as a Globus
identity proivder, for use with this endpoint.

Once configured, you will be able to authenticate users from the associated domain.

== EXAMPLE

globus-connect-server oidc register

""",
)
@click.option("--display-name", help="Display name for the OIDC server", required=True)
@click.option(
    "--discovery-url",
    metavar="URL",
    help=(
        "The OpenID Connect discovery URL for the server.  The domain of the url "
        "will be the main domain of the IdP and must be asserted by the OIDC server. "
        "Example: https://id.example.com/.well-known/openid-configuration"
    ),
    required=True,
)
@click.option(
    "--domain",
    metavar="DOMAIN",
    help=(
        "An alternate domain asserted by the OIDC server.  It is not necessary to "
        "pass the domain from the 'discovery-url'.  May be passed multiple times."
    ),
    callback=verify_domain_name_option,
    multiple=True,
    default=[],
)
@click.option(
    "--client-id",
    metavar="CLIENT-ID",
    help=(
        "You must configure Globus Auth as a client on your OIDC Server.  This is the "
        "Client ID of that that client configuration, which Globus will use to "
        "authenticate with your OIDC server. "
        "This is not the Globus client-id of your endpoint. "
    ),
    required=True,
)
@click.option(
    "--client-secret",
    metavar="SECRET",
    help=(
        "You must configure Globus Auth as a client on your OIDC Server.  This is the "
        "Client secret for the Client ID that Globus will use to authenticate "
        "with your OIDC server.  This is not the Globus client-secret of your endpoint."
    ),
    required=True,
)
@click.option(
    "--username-claim",
    help=(
        "Identity provider claim that maps to the user name. In general this "
        "should be 'preferred_username'."
    ),
    required=True,
)
@click.option(
    "--id-claim",
    help=(
        "Identity provider claim that maps to the immutable ID. In general "
        "this should be 'sub'."
    ),
    required=True,
)
@click.option(
    "--support-contact", help="Support contact name for the OIDC server", required=True
)
@click.option(
    "--support-email",
    help="Support contact email address for the OIDC server",
    required=True,
)
@click.pass_context
def oidc_register(
    ctx,
    display_name,
    discovery_url,
    domain,
    support_contact,
    support_email,
    client_id,
    client_secret,
    username_claim,
    id_claim,
) -> None:
    if os.getuid() != 0:
        click.echo("This command must be run as the root user", err=True)
        ctx.exit(1)

    app = make_app(app_config="globus.manager.api.config.oidc_cli")
    models.db.init_app(app)

    with app.app_context():
        models.db.create_all()
        models.db.session.commit()

        # Needed by AppSync
        for conf in models.Configuration.query:
            if conf.name == "node_key":
                value = NodeKey(**json.loads(conf.value))
            else:
                value = conf.value
            app.config[conf.name.upper()] = value

        # Make sure `node setup` has been run
        if "CLIENT_ID" not in app.config:
            click.echo("This node has not been set up.", err=True)
            click.get_current_context().exit(1)

        try:
            # Confirm that the loaded configuration contains the client ID and secret,
            # otherwise this node has not been configured yet. Can also be checked in app.py:266.
            if not (app.config.keys() > {"CLIENT_ID", "CLIENT_SECRET"}):
                raise Exception(
                    "GCS node has not been configured. Please run 'node setup' "
                    "or run this command on an existing node."
                )

            # Load the keychain
            load_keychain()

            # Get the endpoint
            endpoint_model = models.Endpoint.query.one()
            if endpoint_model.has_oidc_server:
                click.echo(
                    "There is already a managed OIDC configuration for this endpoint, "
                    "please use the `oidc update` command to update parameters."
                )
                return
            elif not endpoint_model.has_oidc_registration:
                current_app.logger.debug("pre init")
                endpoint_model.oidc_registration = models.OIDCRegistration(
                    doc={
                        "DATA_TYPE": "oidc_registration#1.0.0",
                        "display_name": display_name,
                        "discovery_url": discovery_url,
                        "domains": domain,
                        "support_contact": support_contact,
                        "support_email": support_email,
                        "username_claim": username_claim,
                        "id_claim": id_claim,
                        "oidc_client_id": client_id,
                        "oidc_client_secret": client_secret,
                    }
                )
                current_app.logger.debug("init done")
                endpoint_model.refresh_configuration_hash()
                current_app.appsync_client.update_object(endpoint_model)
                models.db.session.add(endpoint_model)
                models.db.session.commit()

            try:
                idp_data = endpoint_model.oidc_registration.auth_identity_provider_data
                # FIXME: I can't tell if the idp data corresponds
                # to a registration for this domain for a different client ID
                # that got nuked
            except AttributeError:
                idp_data = None

            if not idp_data:
                # If the OIDC IDP hasn't been registered yet, we will try to do
                # so
                resp = None
                # Make sure the autodiscovery URL is reachable
                for _tries in range(10):
                    time.sleep(1)
                    try:
                        resp = requests.get(
                            endpoint_model.oidc_registration.discovery_url
                        )
                        if resp.status_code == 200:
                            break
                    except Exception as e:
                        err = "Error contacting discovery url: {}".format(str(e))
                        if _tries < 9:
                            current_app.logger.info(err)
                        else:
                            click.echo(err, err=True)
                            click.get_current_context().exit(1)

                if resp and resp.status_code == 200:
                    try:
                        endpoint_model.oidc_registration.register_idp()
                    except Exception as e:
                        click.echo("Error registering IdP: {}".format(str(e)), err=True)
                        click.get_current_context().exit(1)

                    endpoint_model.refresh_configuration_hash()
                    current_app.appsync_client.update_object(endpoint_model)
                    click.echo("The OIDC server has been registered for this endpoint.")
                    return

        except Exception as e:
            click.echo("Error while creating OIDC endpoint: {}".format(str(e)))
            click.get_current_context().exit(1)
